# -*- coding: utf-8 -*-
"""
Tests for color difference (Delta E) equations.
"""

import unittest

from colorir.colormath.color_diff import (
    delta_e_cie1976,
    delta_e_cie1994,
    delta_e_cie2000,
    delta_e_cmc,
)
from colorir.colormath.color_objects import LabColor, sRGBColor


class DeltaETestCase(unittest.TestCase):
    def setUp(self):
        self.color1 = LabColor(lab_l=0.9, lab_a=16.3, lab_b=-2.22)
        self.color2 = LabColor(lab_l=0.7, lab_a=14.2, lab_b=-1.80)

    def test_cie2000_accuracy(self):
        result = delta_e_cie2000(self.color1, self.color2)
        expected = 1.523
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CIE2000 formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie2000_accuracy_2(self):
        """
        Follow a different execution path based on variable values.
        """

        # These values are from ticket 8 in regards to a CIE2000 bug.
        c1 = LabColor(lab_l=32.8911, lab_a=-53.0107, lab_b=-43.3182)
        c2 = LabColor(lab_l=77.1797, lab_a=25.5928, lab_b=17.9412)
        result = delta_e_cie2000(c1, c2)
        expected = 78.772
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CIE2000 formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie2000_accuracy_3(self):
        """
        Reference:
        "The CIEDE2000 Color-Difference Formula: Implementation Notes,
        Supplementary Test Data, and Mathematical Observations,", G. Sharma,
        W. Wu, E. N. Dalal, submitted to Color Research and Application,
        January 2004. http://www.ece.rochester.edu/~gsharma/ciede2000/
        """

        color1 = (
            LabColor(lab_l=50.0000, lab_a=2.6772, lab_b=-79.7751),
            LabColor(lab_l=50.0000, lab_a=3.1571, lab_b=-77.2803),
            LabColor(lab_l=50.0000, lab_a=2.8361, lab_b=-74.0200),
            LabColor(lab_l=50.0000, lab_a=-1.3802, lab_b=-84.2814),
            LabColor(lab_l=50.0000, lab_a=-1.1848, lab_b=-84.8006),
            LabColor(lab_l=50.0000, lab_a=-0.9009, lab_b=-85.5211),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=-1.0000, lab_b=2.0000),
            LabColor(lab_l=50.0000, lab_a=2.4900, lab_b=-0.0010),
            LabColor(lab_l=50.0000, lab_a=2.4900, lab_b=-0.0010),
            LabColor(lab_l=50.0000, lab_a=2.4900, lab_b=-0.0010),
            LabColor(lab_l=50.0000, lab_a=2.4900, lab_b=-0.0010),
            LabColor(lab_l=50.0000, lab_a=-0.0010, lab_b=2.4900),
            LabColor(lab_l=50.0000, lab_a=-0.0010, lab_b=2.4900),
            LabColor(lab_l=50.0000, lab_a=-0.0010, lab_b=2.4900),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=2.5000, lab_b=0.0000),
            LabColor(lab_l=60.2574, lab_a=-34.0099, lab_b=36.2677),
            LabColor(lab_l=63.0109, lab_a=-31.0961, lab_b=-5.8663),
            LabColor(lab_l=61.2901, lab_a=3.7196, lab_b=-5.3901),
            LabColor(lab_l=35.0831, lab_a=-44.1164, lab_b=3.7933),
            LabColor(lab_l=22.7233, lab_a=20.0904, lab_b=-46.6940),
            LabColor(lab_l=36.4612, lab_a=47.8580, lab_b=18.3852),
            LabColor(lab_l=90.8027, lab_a=-2.0831, lab_b=1.4410),
            LabColor(lab_l=90.9257, lab_a=-0.5406, lab_b=-0.9208),
            LabColor(lab_l=6.7747, lab_a=-0.2908, lab_b=-2.4247),
            LabColor(lab_l=2.0776, lab_a=0.0795, lab_b=-1.1350),
        )
        color2 = (
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-82.7485),
            LabColor(lab_l=50.0000, lab_a=-1.0000, lab_b=2.0000),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=-2.4900, lab_b=0.0009),
            LabColor(lab_l=50.0000, lab_a=-2.4900, lab_b=0.0010),
            LabColor(lab_l=50.0000, lab_a=-2.4900, lab_b=0.0011),
            LabColor(lab_l=50.0000, lab_a=-2.4900, lab_b=0.0012),
            LabColor(lab_l=50.0000, lab_a=0.0009, lab_b=-2.4900),
            LabColor(lab_l=50.0000, lab_a=0.0010, lab_b=-2.4900),
            LabColor(lab_l=50.0000, lab_a=0.0011, lab_b=-2.4900),
            LabColor(lab_l=50.0000, lab_a=0.0000, lab_b=-2.5000),
            LabColor(lab_l=73.0000, lab_a=25.0000, lab_b=-18.0000),
            LabColor(lab_l=61.0000, lab_a=-5.0000, lab_b=29.0000),
            LabColor(lab_l=56.0000, lab_a=-27.0000, lab_b=-3.0000),
            LabColor(lab_l=58.0000, lab_a=24.0000, lab_b=15.0000),
            LabColor(lab_l=50.0000, lab_a=3.1736, lab_b=0.5854),
            LabColor(lab_l=50.0000, lab_a=3.2972, lab_b=0.0000),
            LabColor(lab_l=50.0000, lab_a=1.8634, lab_b=0.5757),
            LabColor(lab_l=50.0000, lab_a=3.2592, lab_b=0.3350),
            LabColor(lab_l=60.4626, lab_a=-34.1751, lab_b=39.4387),
            LabColor(lab_l=62.8187, lab_a=-29.7946, lab_b=-4.0864),
            LabColor(lab_l=61.4292, lab_a=2.2480, lab_b=-4.9620),
            LabColor(lab_l=35.0232, lab_a=-40.0716, lab_b=1.5901),
            LabColor(lab_l=23.0331, lab_a=14.9730, lab_b=-42.5619),
            LabColor(lab_l=36.2715, lab_a=50.5065, lab_b=21.2231),
            LabColor(lab_l=91.1528, lab_a=-1.6435, lab_b=0.0447),
            LabColor(lab_l=88.6381, lab_a=-0.8985, lab_b=-0.7239),
            LabColor(lab_l=5.8714, lab_a=-0.0985, lab_b=-2.2286),
            LabColor(lab_l=0.9033, lab_a=-0.0636, lab_b=-0.5514),
        )
        diff = (
            2.0425,
            2.8615,
            3.4412,
            1.0000,
            1.0000,
            1.0000,
            2.3669,
            2.3669,
            7.1792,
            7.1792,
            7.2195,
            7.2195,
            4.8045,
            4.8045,
            4.7461,
            4.3065,
            27.1492,
            22.8977,
            31.9030,
            19.4535,
            1.0000,
            1.0000,
            1.0000,
            1.0000,
            1.2644,
            1.2630,
            1.8731,
            1.8645,
            2.0373,
            1.4146,
            1.4441,
            1.5381,
            0.6377,
            0.9082,
        )
        for l_set in zip(color1, color2, diff):
            result = delta_e_cie2000(l_set[0], l_set[1])
            expected = l_set[2]
            self.assertAlmostEqual(
                result,
                expected,
                4,
                "DeltaE CIE2000 formula error. "
                "Got %.4f, expected %.4f (diff: %.4f)."
                % (result, expected, result - expected),
            )

    def test_cie1994_negative_square_root(self):
        """
        Tests against a case where a negative square root in the delta_H
        calculation could happen.
        """

        standard = LabColor(lab_l=0.9, lab_a=1, lab_b=1)
        sample = LabColor(lab_l=0.7, lab_a=0, lab_b=0)
        delta_e_cie1994(standard, sample)

    def test_cmc_negative_square_root(self):
        """
        Tests against a case where a negative square root in the delta_H
        calculation could happen.
        """

        standard = LabColor(lab_l=0.9, lab_a=1, lab_b=1)
        sample = LabColor(lab_l=0.7, lab_a=0, lab_b=0)
        delta_e_cmc(standard, sample)

    # noinspection PyArgumentEqualDefault
    def test_cmc_accuracy(self):
        # Test 2:1
        result = delta_e_cmc(self.color1, self.color2, pl=2, pc=1)
        expected = 1.443
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CMC (2:1) formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

        # Test against 1:1 as well
        result = delta_e_cmc(self.color1, self.color2, pl=1, pc=1)
        expected = 1.482
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CMC (1:1) formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

        # Testing atan H behavior.
        atan_color1 = LabColor(lab_l=69.417, lab_a=-12.612, lab_b=-11.271)
        atan_color2 = LabColor(lab_l=83.386, lab_a=39.426, lab_b=-17.525)
        result = delta_e_cmc(atan_color1, atan_color2)
        expected = 44.346
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CMC Atan test formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie1976_accuracy(self):
        result = delta_e_cie1976(self.color1, self.color2)
        expected = 2.151
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CIE1976 formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie1994_accuracy_graphic_arts(self):
        result = delta_e_cie1994(self.color1, self.color2)
        expected = 1.249
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CIE1994 (graphic arts) formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie1994_accuracy_textiles(self):
        result = delta_e_cie1994(self.color1, self.color2, K_1=0.048, K_2=0.014, K_L=2)
        expected = 1.204
        self.assertAlmostEqual(
            result,
            expected,
            3,
            "DeltaE CIE1994 (textiles) formula error. "
            "Got %.3f, expected %.3f (diff: %.3f)."
            % (result, expected, result - expected),
        )

    def test_cie1994_domain_error(self):
        # These values are from ticket 98 in regards to a CIE1995
        # domain error exception being raised.
        c1 = LabColor(lab_l=50, lab_a=0, lab_b=0)
        c2 = LabColor(lab_l=50, lab_a=-1, lab_b=2)
        try:
            delta_e_cie1994(c1, c2)
        except ValueError:
            self.fail("DeltaE CIE1994 domain error.")

    def test_non_lab_color(self):
        other_color = sRGBColor(1.0, 0.5, 0.3)
        self.assertRaises(ValueError, delta_e_cie2000, self.color1, other_color)



# -*- coding: utf-8 -*-
"""
Tests for color difference (Delta E) equations.
"""

import unittest

from colorir.colormath.color_objects import XYZColor


# noinspection PyPep8Naming
class chromaticAdaptationTestCase(unittest.TestCase):
    def setUp(self):
        self.color = XYZColor(0.5, 0.4, 0.1, illuminant="C")

    def test_adaptation_c_to_d65(self):
        self.color.apply_adaptation(target_illuminant="D65")
        self.assertAlmostEqual(
            self.color.xyz_x, 0.491, 3, "C to D65 adaptation failed: X coord"
        )
        self.assertAlmostEqual(
            self.color.xyz_y, 0.400, 3, "C to D65 adaptation failed: Y coord"
        )
        self.assertAlmostEqual(
            self.color.xyz_z, 0.093, 3, "C to D65 adaptation failed: Z coord"
        )
        self.assertEqual(
            self.color.illuminant,
            "d65",
            "C to D65 adaptation failed: Illuminant transfer",
        )



# -*- coding: utf-8 -*-
"""
Various tests for color objects.
"""

import unittest

from colorir.colormath.color_conversions import convert_color
from colorir.colormath.color_objects import (
    SpectralColor,
    XYZColor,
    xyYColor,
    LabColor,
    LuvColor,
    LCHabColor,
    LCHuvColor,
    sRGBColor,
    HSLColor,
    HSVColor,
    CMYColor,
    CMYKColor,
    AdobeRGBColor,
    AppleRGBColor,
    IPTColor,
)


class BaseColorConversionTest(unittest.TestCase):
    """
    All color conversion tests should inherit from this class. Has some
    convenience methods for re-use.
    """

    # noinspection PyPep8Naming
    def assertColorMatch(self, conv, std):
        """
        Checks a converted color against an expected standard.

        :param conv: The converted color object.
        :param std: The object to use as a standard for comparison.
        """

        self.assertEqual(conv.__class__, std.__class__)
        attribs = std.VALUES
        for attrib in attribs:
            conv_value = getattr(conv, attrib)
            std_value = getattr(std, attrib)
            self.assertAlmostEqual(
                conv_value,
                std_value,
                3,
                "%s is %s, expected %s" % (attrib, conv_value, std_value),
            )


class SpectralConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        """
        While it is possible to specify the entire spectral color using
        positional arguments, set this thing up with keywords for the ease of
        manipulation.
        """

        color = SpectralColor(
            spec_380nm=0.0600,
            spec_390nm=0.0600,
            spec_400nm=0.0641,
            spec_410nm=0.0654,
            spec_420nm=0.0645,
            spec_430nm=0.0605,
            spec_440nm=0.0562,
            spec_450nm=0.0543,
            spec_460nm=0.0537,
            spec_470nm=0.0541,
            spec_480nm=0.0559,
            spec_490nm=0.0603,
            spec_500nm=0.0651,
            spec_510nm=0.0680,
            spec_520nm=0.0705,
            spec_530nm=0.0736,
            spec_540nm=0.0772,
            spec_550nm=0.0809,
            spec_560nm=0.0870,
            spec_570nm=0.0990,
            spec_580nm=0.1128,
            spec_590nm=0.1251,
            spec_600nm=0.1360,
            spec_610nm=0.1439,
            spec_620nm=0.1511,
            spec_630nm=0.1590,
            spec_640nm=0.1688,
            spec_650nm=0.1828,
            spec_660nm=0.1996,
            spec_670nm=0.2187,
            spec_680nm=0.2397,
            spec_690nm=0.2618,
            spec_700nm=0.2852,
            spec_710nm=0.2500,
            spec_720nm=0.2400,
            spec_730nm=0.2300,
        )
        self.color = color

    def test_conversion_to_xyz(self):
        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.115, 0.099, 0.047))

    def test_conversion_to_xyz_with_negatives(self):
        """
        This has negative spectral values, which should never happen. Just
        clamp these to 0.0 instead of running into the domain errors. A badly
        or uncalibrated spectro can sometimes report negative values.
        """

        self.color.spec_530nm = -0.0736
        # TODO: Convert here.

    def test_convert_to_self(self):
        same_color = convert_color(self.color, SpectralColor)
        self.assertEqual(self.color, same_color)


class XYZConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = XYZColor(0.1, 0.2, 0.3)

    def test_conversion_to_xyy(self):
        xyy = convert_color(self.color, xyYColor)
        self.assertColorMatch(xyy, xyYColor(0.167, 0.333, 0.200))

    def test_conversion_to_lab(self):
        lab = convert_color(self.color, LabColor)
        self.assertColorMatch(lab, LabColor(51.837, -57.486, -25.780))

    def test_conversion_to_rgb(self):
        # Picked a set of XYZ coordinates that would return a good RGB value.
        self.color = XYZColor(0.300, 0.200, 0.300)
        rgb = convert_color(self.color, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.715, 0.349, 0.663))

    def test_conversion_to_apple_rgb(self):
        self.color = XYZColor(0.0157, 0.0191, 0.0331)
        rgb = convert_color(self.color, AppleRGBColor)
        self.assertColorMatch(rgb, AppleRGBColor(0.0411, 0.1214, 0.1763))

    def test_conversion_to_adobe_rgb(self):
        self.color = XYZColor(0.2553, 0.1125, 0.0011)
        rgb = convert_color(self.color, AdobeRGBColor)
        # This ends up getting clamped.
        self.assertColorMatch(rgb, AdobeRGBColor(0.6828, 0.0, 0.0))

    def test_conversion_to_luv(self):
        luv = convert_color(self.color, LuvColor)
        self.assertColorMatch(luv, LuvColor(51.837, -73.561, -25.657))

    def test_conversion_to_ipt(self):
        self.color.set_illuminant("D65")
        ipt = convert_color(self.color, IPTColor)
        self.assertColorMatch(ipt, IPTColor(0.5063, -0.3183, -0.1160))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, XYZColor)
        self.assertEqual(self.color, same_color)


# noinspection PyPep8Naming
class xyYConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = xyYColor(0.167, 0.333, 0.200)

    def test_conversion_to_xyz(self):
        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.100, 0.200, 0.300))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, xyYColor)
        self.assertEqual(self.color, same_color)


class LabConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = LabColor(1.807, -3.749, -2.547)

    def test_conversion_to_xyz(self):
        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.001, 0.002, 0.003))

    def test_conversion_to_lchab(self):
        lch = convert_color(self.color, LCHabColor)
        self.assertColorMatch(lch, LCHabColor(1.807, 4.532, 214.191))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, LabColor)
        self.assertEqual(self.color, same_color)


class LuvConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = LuvColor(1.807, -2.564, -0.894)

    def test_conversion_to_xyz(self):
        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.001, 0.002, 0.003))

    def test_conversion_to_lchuv(self):
        lch = convert_color(self.color, LCHuvColor)
        self.assertColorMatch(lch, LCHuvColor(1.807, 2.715, 199.222))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, LuvColor)
        self.assertEqual(self.color, same_color)


# noinspection PyAttributeOutsideInit,PyPep8Naming
class LCHabConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = LCHabColor(1.807, 4.532, 214.191)

    def test_conversion_to_lab(self):
        lab = convert_color(self.color, LabColor)
        self.assertColorMatch(lab, LabColor(1.807, -3.749, -2.547))

    def test_conversion_to_rgb_zero_div(self):
        """
        The formula I grabbed for LCHuv to XYZ had a zero division error in it
        if the L coord was 0. Also check against LCHab in case.
        """

        lchab = LCHabColor(0.0, 0.0, 0.0)
        rgb = convert_color(lchab, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.0, 0.0, 0.0))

    def test_to_rgb_domain_error(self):
        """
        Tests for a bug resulting in a domain error with LCH->Adobe RGB.

        See: https://github.com/gtaylor/python-colormath/issues/49
        """

        lchab = LCHabColor(40.0, 104.0, 40.0)
        _rgb = convert_color(lchab, AdobeRGBColor)  # noqa

    def test_convert_to_self(self):
        same_color = convert_color(self.color, LCHabColor)
        self.assertEqual(self.color, same_color)


class LCHuvConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = LCHuvColor(1.807, 2.715, 199.228)

    def test_conversion_to_luv(self):
        luv = convert_color(self.color, LuvColor)
        self.assertColorMatch(luv, LuvColor(1.807, -2.564, -0.894))

    def test_conversion_to_rgb_zero_div(self):
        """
        The formula I grabbed for LCHuv to XYZ had a zero division error in it
        if the L coord was 0. Check against that here.

        Issue #13 in the Google Code tracker.
        """

        lchuv = LCHuvColor(0.0, 0.0, 0.0)
        rgb = convert_color(lchuv, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.0, 0.0, 0.0))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, LCHuvColor)
        self.assertEqual(self.color, same_color)


class RGBConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = sRGBColor(0.482, 0.784, 0.196)

    def test_channel_clamping(self):
        high_r = sRGBColor(1.482, 0.2, 0.3)
        self.assertEqual(high_r.clamped_rgb_r, 1.0)
        self.assertEqual(high_r.clamped_rgb_g, high_r.rgb_g)
        self.assertEqual(high_r.clamped_rgb_b, high_r.rgb_b)

        low_r = sRGBColor(-0.1, 0.2, 0.3)
        self.assertEqual(low_r.clamped_rgb_r, 0.0)
        self.assertEqual(low_r.clamped_rgb_g, low_r.rgb_g)
        self.assertEqual(low_r.clamped_rgb_b, low_r.rgb_b)

        high_g = sRGBColor(0.2, 1.482, 0.3)
        self.assertEqual(high_g.clamped_rgb_r, high_g.rgb_r)
        self.assertEqual(high_g.clamped_rgb_g, 1.0)
        self.assertEqual(high_g.clamped_rgb_b, high_g.rgb_b)

        low_g = sRGBColor(0.2, -0.1, 0.3)
        self.assertEqual(low_g.clamped_rgb_r, low_g.rgb_r)
        self.assertEqual(low_g.clamped_rgb_g, 0.0)
        self.assertEqual(low_g.clamped_rgb_b, low_g.rgb_b)

        high_b = sRGBColor(0.1, 0.2, 1.482)
        self.assertEqual(high_b.clamped_rgb_r, high_b.rgb_r)
        self.assertEqual(high_b.clamped_rgb_g, high_b.rgb_g)
        self.assertEqual(high_b.clamped_rgb_b, 1.0)

        low_b = sRGBColor(0.1, 0.2, -0.1)
        self.assertEqual(low_b.clamped_rgb_r, low_b.rgb_r)
        self.assertEqual(low_b.clamped_rgb_g, low_b.rgb_g)
        self.assertEqual(low_b.clamped_rgb_b, 0.0)

    def test_to_xyz_and_back(self):
        xyz = convert_color(self.color, XYZColor)
        rgb = convert_color(xyz, sRGBColor)
        self.assertColorMatch(rgb, self.color)

    def test_conversion_to_hsl_max_r(self):
        color = sRGBColor(255, 123, 50, is_upscaled=True)
        hsl = convert_color(color, HSLColor)
        self.assertColorMatch(hsl, HSLColor(21.366, 1.000, 0.598))

    def test_conversion_to_hsl_max_g(self):
        color = sRGBColor(123, 255, 50, is_upscaled=True)
        hsl = convert_color(color, HSLColor)
        self.assertColorMatch(hsl, HSLColor(98.634, 1.000, 0.598))

    def test_conversion_to_hsl_max_b(self):
        color = sRGBColor(0.482, 0.482, 1.0)
        hsl = convert_color(color, HSLColor)
        self.assertColorMatch(hsl, HSLColor(240.000, 1.000, 0.741))

    def test_conversion_to_hsl_gray(self):
        color = sRGBColor(0.482, 0.482, 0.482)
        hsl = convert_color(color, HSLColor)
        self.assertColorMatch(hsl, HSLColor(0.000, 0.000, 0.482))

    def test_conversion_to_hsv(self):
        hsv = convert_color(self.color, HSVColor)
        self.assertColorMatch(hsv, HSVColor(90.816, 0.750, 0.784))

    def test_conversion_to_cmy(self):
        cmy = convert_color(self.color, CMYColor)
        self.assertColorMatch(cmy, CMYColor(0.518, 0.216, 0.804))

    def test_srgb_conversion_to_xyz_d50(self):
        """
        sRGB's native illuminant is D65. Test the XYZ adaptations by setting
        a target illuminant to something other than D65.
        """

        xyz = convert_color(self.color, XYZColor, target_illuminant="D50")
        self.assertColorMatch(xyz, XYZColor(0.313, 0.460, 0.082))

    def test_srgb_conversion_to_xyz_d65(self):
        """
        sRGB's native illuminant is D65. This is a straightforward conversion.
        """

        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.294, 0.457, 0.103))

    def test_adobe_conversion_to_xyz_d65(self):
        """
        Adobe RGB's native illuminant is D65, like sRGB's. However, sRGB uses
        different conversion math that uses gamma, so test the alternate logic
        route for non-sRGB RGB colors.
        """

        adobe = AdobeRGBColor(0.482, 0.784, 0.196)
        xyz = convert_color(adobe, XYZColor)
        self.assertColorMatch(xyz, XYZColor(0.230, 0.429, 0.074))

    def test_conversion_through_rgb(self):
        """
        Make sure our convenience RGB tracking feature is working. For example,
        going from XYZ->HSL via Adobe RGB, then taking that HSL object and
        converting back to XYZ should also use Adobe RGB (instead of the
        default of sRGB).
        """

        xyz = convert_color(self.color, XYZColor)
        hsl = convert_color(xyz, HSLColor, through_rgb_type=AdobeRGBColor)
        # Notice how we don't have to pass through_rgb_type explicitly.
        xyz2 = convert_color(hsl, XYZColor)
        self.assertColorMatch(xyz, xyz2)

    def test_adobe_conversion_to_xyz_d50(self):
        """
        Adobe RGB's native illuminant is D65, so an adaptation matrix is
        involved here. However, the math for sRGB and all other RGB types is
        different, so test all of the other types with an adaptation matrix
        here.
        """

        adobe = AdobeRGBColor(0.482, 0.784, 0.196)
        xyz = convert_color(adobe, XYZColor, target_illuminant="D50")
        self.assertColorMatch(xyz, XYZColor(0.247, 0.431, 0.060))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, sRGBColor)
        self.assertEqual(self.color, same_color)

    def test_get_rgb_hex(self):
        hex_str = self.color.get_rgb_hex()
        self.assertEqual(hex_str, "#7bc832", "sRGB to hex conversion failed")

    def test_set_from_rgb_hex(self):
        rgb = sRGBColor.new_from_rgb_hex("#7bc832")
        self.assertColorMatch(rgb, sRGBColor(0.482, 0.784, 0.196))


class HSLConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = HSLColor(200.0, 0.400, 0.500)

    def test_conversion_to_rgb(self):
        rgb = convert_color(self.color, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.300, 0.567, 0.700))
        # Make sure this converts to AdobeRGBColor instead of sRGBColor.
        adobe_rgb = convert_color(self.color, AdobeRGBColor)
        self.assertIsInstance(adobe_rgb, AdobeRGBColor)

    def test_convert_to_self(self):
        same_color = convert_color(self.color, HSLColor)
        self.assertEqual(self.color, same_color)


class HSVConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = HSVColor(91.0, 0.750, 0.784)

    def test_conversion_to_rgb(self):
        rgb = convert_color(self.color, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.480, 0.784, 0.196))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, HSVColor)
        self.assertEqual(self.color, same_color)


class CMYConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = CMYColor(0.518, 0.216, 0.804)

    def test_conversion_to_cmyk(self):
        cmyk = convert_color(self.color, CMYKColor)
        self.assertColorMatch(cmyk, CMYKColor(0.385, 0.000, 0.750, 0.216))

    def test_conversion_to_rgb(self):
        rgb = convert_color(self.color, sRGBColor)
        self.assertColorMatch(rgb, sRGBColor(0.482, 0.784, 0.196))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, CMYColor)
        self.assertEqual(self.color, same_color)


class CMYKConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = CMYKColor(0.385, 0.000, 0.750, 0.216)

    def test_conversion_to_cmy(self):
        cmy = convert_color(self.color, CMYColor)
        self.assertColorMatch(cmy, CMYColor(0.518, 0.216, 0.804))

    def test_convert_to_self(self):
        same_color = convert_color(self.color, CMYKColor)
        self.assertEqual(self.color, same_color)


class IPTConversionTestCase(BaseColorConversionTest):
    def setUp(self):
        self.color = IPTColor(0.5, 0.5, 0.5)

    def test_convert_so_self(self):
        same_color = convert_color(self.color, IPTColor)
        self.assertEqual(self.color, same_color)

    def test_convert_to_XYZ(self):
        xyz = convert_color(self.color, XYZColor)
        self.assertColorMatch(
            xyz, XYZColor(0.4497, 0.2694, 0.0196, illuminant="d65", observer="2")
        )

    def test_consistency(self):
        xyz = convert_color(self.color, XYZColor)
        same_color = convert_color(xyz, IPTColor)
        self.assertColorMatch(self.color, same_color)

    def test_illuminant_guard(self):
        xyz = XYZColor(1, 1, 1, illuminant="d50", observer="2")

        def _ipt_conversion():
            return convert_color(xyz, IPTColor)

        self.assertRaises(ValueError, _ipt_conversion)

    def test_observer_guard(self):
        xyz = XYZColor(1, 1, 1, illuminant="d65", observer="10")

        def _ipt_conversion():
            return convert_color(xyz, IPTColor)

        self.assertRaises(ValueError, _ipt_conversion)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2014, Michael Mauderer, University of St Andrews
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of St Andrews nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division

import logging
import numpy

logger = logging.getLogger(__name__)


class Nayatani95(object):
    """
    **References**

    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.
    *   Nayatani, Y., Sobagaki, H., & Yano, K. H. T. (1995). Lightness dependency of
        chroma scales of a nonlinear color-appearance model and its latest formulation.
        *Color Research & Application*, 20(3), 156-167.
    """

    @property
    def hue_angle(self):
        """
        Predicted hue angle :math:`\\theta`.
        """
        return self._hue_angle

    @property
    def chroma(self):
        """
        Predicted chroma :math:`C`.
        """
        return self._chroma

    @property
    def saturation(self):
        """
        Predicted saturation :math:`S`.
        """
        return self._saturation

    @property
    def brightness(self):
        """
        Predicted brightness :math:`B_r`.
        """
        return self._brightness

    @property
    def colorfulness(self):
        """
        Predicted colorfulness :math:`M`.
        """
        return self._colorfulness

    def __init__(self, x, y, z, x_n, y_n, z_n, y_ob, e_o, e_or, n=1):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_n: X value of reference white :math:`X_n`.
        :param y_n: Y value of reference white :math:`Y_n`.
        :param z_n: Z value of reference white :math:`Z_n`.
        :param y_ob: Luminance factor of achromatic background as percentage :math:`Y_o`.
                     Required to be larger than 0.18.
        :param e_o: Illuminance of the viewing field :math:`E_o` in lux.
        :param e_or: Normalising illuminance :math:`E_or` in lux.
        :param n: Noise term :math:`n`.
        """

        if numpy.any(y_ob <= 0.18):
            raise ValueError("y_ob hast be greater than 0.18.")

        l_o = y_ob * e_o / (100 * numpy.pi)
        l_or = y_ob * e_or / (100 * numpy.pi)
        logger.debug("L_o: {}".format(l_o))
        logger.debug("L_or: {}".format(l_or))

        x_o = x_n / (x_n + y_n + z_n)
        y_o = y_n / (x_n + y_n + z_n)
        logger.debug("x_o: {}".format(x_o))
        logger.debug("y_o: {}".format(y_o))

        xi = (0.48105 * x_o + 0.78841 * y_o - 0.08081) / y_o
        eta = (-0.27200 * x_o + 1.11962 * y_o + 0.04570) / y_o
        zeta = (0.91822 * (1 - x_o - y_o)) / y_o
        logger.debug("xi: {}".format(xi))
        logger.debug("eta: {}".format(eta))
        logger.debug("zeta: {}".format(zeta))

        r_0, g_0, b_0 = rgb_0 = ((y_ob * e_o) / (100 * numpy.pi)) * numpy.array(
            [xi, eta, zeta]
        )
        logger.debug("rgb_0: {}".format(rgb_0))

        r, g, b, = rgb = self.xyz_to_rgb(numpy.array([x, y, z]))
        logger.debug("rgb: {}".format(rgb))

        e_r = self._compute_scaling_coefficient(r, xi)
        logger.debug("e(R): {}".format(e_r))
        e_g = self._compute_scaling_coefficient(g, eta)
        logger.debug("e(G): {}".format(e_g))

        beta_r = self._beta_1(r_0)
        logger.debug("beta1(rho): {}".format(beta_r))
        beta_g = self._beta_1(g_0)
        logger.debug("beta1(eta): {}".format(beta_g))
        beta_b = self._beta_2(b_0)
        logger.debug("beta2(zeta): {}".format(beta_b))

        beta_l = self._beta_1(l_or)
        logger.debug("beta1(L_or): {}".format(beta_l))

        # Opponent Color Dimension
        self._achromatic_response = (
            (2 / 3) * beta_r * e_r * numpy.log10((r + n) / (20 * xi + n))
        )
        self._achromatic_response += (
            (1 / 3) * beta_g * e_g * numpy.log10((g + n) / (20 * eta + n))
        )
        self._achromatic_response *= 41.69 / beta_l
        logger.debug("Q: {}".format(self._achromatic_response))

        self._tritanopic_response = (
            (1 / 1) * beta_r * numpy.log10((r + n) / (20 * xi + n))
        )
        self._tritanopic_response += (
            -(12 / 11) * beta_g * numpy.log10((g + n) / (20 * eta + n))
        )
        self._tritanopic_response += (
            (1 / 11) * beta_b * numpy.log10((b + n) / (20 * zeta + n))
        )
        logger.debug("t: {}".format(self._tritanopic_response))

        self._protanopic_response = (
            (1 / 9) * beta_r * numpy.log10((r + n) / (20 * xi + n))
        )
        self._protanopic_response += (
            (1 / 9) * beta_g * numpy.log10((g + n) / (20 * eta + n))
        )
        self._protanopic_response += (
            -(2 / 9) * beta_b * numpy.log10((b + n) / (20 * zeta + n))
        )
        logger.debug("p: {}".format(self._protanopic_response))

        # Brightness
        self._brightness = (50 / beta_l) * (
            (2 / 3) * beta_r + (1 / 3) * beta_g
        ) + self._achromatic_response

        self._brightness_ideal_white = (
            (2 / 3) * beta_r * 1.758 * numpy.log10((100 * xi + n) / (20 * xi + n))
        )
        self._brightness_ideal_white += (
            (1 / 3) * beta_g * 1.758 * numpy.log10((100 * eta + n) / (20 * eta + n))
        )
        self._brightness_ideal_white *= 41.69 / beta_l
        self._brightness_ideal_white += (50 / beta_l) * (2 / 3) * beta_r
        self._brightness_ideal_white += (50 / beta_l) * (1 / 3) * beta_g

        # Lightness
        self._lightness_achromatic = self._achromatic_response + 50
        self._lightness_achromatic_normalized = 100 * (
            self._brightness / self._brightness_ideal_white
        )

        # Hue
        hue_angle_rad = numpy.arctan2(
            self._protanopic_response, self._tritanopic_response
        )
        self._hue_angle = ((360 * hue_angle_rad / (2 * numpy.pi)) + 360) % 360
        logger.debug("theta: {}".format(self._hue_angle))

        e_s_theta = self.chromatic_strength(hue_angle_rad)
        logger.debug("E_s(theta): {}".format(e_s_theta))

        # Saturation
        self._saturation_rg = (488.93 / beta_l) * e_s_theta * self._tritanopic_response
        self._saturation_yb = (488.93 / beta_l) * e_s_theta * self._protanopic_response
        logger.debug("S_RG: {}".format(self._saturation_rg))
        logger.debug("S_YB: {}".format(self._saturation_yb))

        self._saturation = numpy.sqrt(
            (self._saturation_rg ** 2) + (self._saturation_yb ** 2)
        )
        logger.debug("S: {}".format(self._saturation))

        # Chroma
        self._chroma_rg = (
            (self._lightness_achromatic / 50) ** 0.7
        ) * self._saturation_rg
        self._chroma_yb = (
            (self._lightness_achromatic / 50) ** 0.7
        ) * self._saturation_yb
        self._chroma = ((self._lightness_achromatic / 50) ** 0.7) * self._saturation
        logger.debug("C: {}".format(self._chroma))

        # Colorfulness
        self._colorfulness_rg = self._chroma_rg * self._brightness_ideal_white / 100
        self._colorfulness_yb = self._chroma_yb * self._brightness_ideal_white / 100
        self._colorfulness = self._chroma * self._brightness_ideal_white / 100

    @staticmethod
    def chromatic_strength(angle):
        result = 0.9394
        result += -0.2478 * numpy.sin(1 * angle)
        result += -0.0743 * numpy.sin(2 * angle)
        result += +0.0666 * numpy.sin(3 * angle)
        result += -0.0186 * numpy.sin(4 * angle)
        result += -0.0055 * numpy.cos(1 * angle)
        result += -0.0521 * numpy.cos(2 * angle)
        result += -0.0573 * numpy.cos(3 * angle)
        result += -0.0061 * numpy.cos(4 * angle)
        return result

    @staticmethod
    def _compute_scaling_coefficient(a, b):
        return numpy.where(a >= (20 * b), 1.758, 1)

    @staticmethod
    def _beta_1(x):
        return (6.469 + 6.362 * (x ** 0.4495)) / (6.469 + (x ** 0.4495))

    @staticmethod
    def _beta_2(x):
        return 0.7844 * (8.414 + 8.091 * (x ** 0.5128)) / (8.414 + (x ** 0.5128))

    xyz_to_rgb_m = numpy.array(
        [[0.40024, 0.70760, -0.08081], [-0.22630, 1.16532, 0.04570], [0, 0, 0.91822]]
    )

    @classmethod
    def xyz_to_rgb(cls, xyz):
        return cls.xyz_to_rgb_m.dot(xyz)


class Hunt(object):
    """
    **References**

    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.
    *   Hunt, R. W. G. (2005). *The reproduction of colour*. 5th Ed., John Wiley & Sons.
    """

    @property
    def hue_angle(self):
        """
        Predicted hue angle :math:`h_s`.
        """
        return self._hue_angle

    @property
    def chroma(self):
        """
        Predicted chroma :math:`C_{94}`.
        """
        return self._chroma

    @property
    def saturation(self):
        """
        Predicted saturation :math:`s`.
        """
        return self._saturation

    @property
    def brightness(self):
        """
        Predicted brightness :math:`Q`.
        """
        return self._brightness

    @property
    def colorfulness(self):
        """
        Predicted colorfulness :math:`M_{94}`.
        """
        return self._colorfulness

    @property
    def lightness(self):
        """
        Predicted colorfulness :math:`J`.
        """
        return self._lightness

    def __init__(
        self,
        x,
        y,
        z,
        x_b,
        y_b,
        z_b,
        x_w,
        y_w,
        z_w,
        l_a,
        n_c,
        n_b,
        l_as=None,
        cct_w=None,
        n_cb=None,
        n_bb=None,
        x_p=None,
        y_p=None,
        z_p=None,
        p=None,
        helson_judd=False,
        discount_illuminant=True,
        s=None,
        s_w=None,
    ):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_b: X value of background :math:`X_b`.
        :param y_b: Y value of background :math:`Y_b`.
        :param z_b: Z value of background :math:`Z_b`.
        :param x_w: X value of reference white :math:`X_W`.
        :param y_w: Y value of reference white :math:`Y_W`.
        :param z_w: Z value of reference white :math:`Z_W`.
        :param l_a: Adapting luminance :math:`L_A`.
        :param n_c: Chromatic surround induction_factor :math:`N_c`.
        :param n_b: Brightness surround induction factor :math:`N_b`.
        :param l_as: Scotopic luminance of the illuminant :math:`L_{AS}`.
                     Will be approximated if not supplied.
        :param cct_w: Correlated color temperature of illuminant :math:`T`.
                      Will be used to approximate l_as if not supplied.
        :param n_cb: Chromatic background induction factor :math:`N_{cb}`.
                     Will be approximated using y_w and y_b if not supplied.
        :param n_bb: Brightness background induction factor :math:`N_{bb}`.
                     Will be approximated using y_w and y_b if not supplied.
        :param x_p: X value of proxima field :math:`X_p`.
                    If not supplied, will be assumed to equal background.
        :param y_p: Y value of proxima field :math:`Y_p`.
                    If not supplied, will be assumed to equal background.
        :param z_p: Z value of proxima field :math:`Z_p`.
                    If not supplied, will be assumed to equal background.
        :param p: Simultaneous contrast/assimilation parameter.
        :param helson_judd: Truth value indicating whether the Heslon-Judd effect
                            should be accounted for.
                            Default False.
        :param discount_illuminant: Truth value whether discount-the-illuminant
                                    should be applied. Default True.
        :param s: Scotopic response to the stimulus.
        :param s_w: Scotopic response for th reference white.
        :raises ValueError: if illegal parameter combination is supplied.
        """
        if x_p is None:
            x_p = x_b
            logger.warn("Approximated x_p with x_b.")
        if y_p is None:
            y_p = y_b
            logger.warn("Approximated y_p with y_b.")
        if z_p is None:
            z_p = y_b
            logger.warn("Approximated z_p with z_b.")

        if n_cb is None:
            n_cb = 0.725 * (y_w / y_b) ** 0.2
            logger.warn("Approximated n_cb.")
        logger.debug("N_cb: {}".format(n_cb))
        if n_bb is None:
            n_bb = 0.725 * (y_w / y_b) ** 0.2
            logger.warn("Approximated n_bb.")
        logger.debug("N_bb: {}".format(n_cb))

        if l_as is None:
            logger.warn("Approximated scotopic luminance.")
            if cct_w is None:
                cct_w = self._get_cct(x_w, y_w, z_w)
                logger.warn("Approximated cct_w: {}".format(cct_w))
            l_as = 2.26 * l_a
            l_as *= ((cct_w / 4000) - 0.4) ** (1 / 3)
        logger.debug("LA_S: {}".format(l_as))

        if (s is None) is not (s_w is None):
            raise ValueError(
                "Either both scotopic responses (s, s_w) need to be supplied or none."
            )
        elif s is None and s_w is None:
            s = y
            s_w = y_w
            logger.warn(
                "Approximated scotopic response to stimulus and reference white."
            )

        if p is None:
            logger.warn(
                "p not supplied. Model will not account for simultaneous "
                "chromatic contrast ."
            )

        xyz = numpy.array([x, y, z])
        logger.debug("XYZ: {}".format(xyz))
        xyz_w = numpy.array([x_w, y_w, z_w])
        logger.debug("XYZ_W: {}".format(xyz_w))
        xyz_b = numpy.array([x_b, y_b, z_b])
        xyz_p = numpy.array([x_p, y_p, z_p])

        k = 1 / (5 * l_a + 1)
        logger.debug("k: {}".format(k))
        # luminance adaptation factor
        f_l = 0.2 * (k ** 4) * (5 * l_a) + 0.1 * ((1 - (k ** 4)) ** 2) * (
            (5 * l_a) ** (1 / 3)
        )
        logger.debug("F_L: {}".format(f_l))

        logger.debug("--- Stimulus RGB adaptation start ----")
        rgb_a = self._adaptation(
            f_l, l_a, xyz, xyz_w, xyz_b, xyz_p, p, helson_judd, discount_illuminant
        )
        logger.debug("--- Stimulus RGB adaptation end ----")
        r_a, g_a, b_a = rgb_a
        logger.debug("RGB_A: {}".format(rgb_a))
        logger.debug("--- White RGB adaptation start ----")
        rgb_aw = self._adaptation(
            f_l, l_a, xyz_w, xyz_w, xyz_b, xyz_p, p, helson_judd, discount_illuminant
        )
        logger.debug("--- White RGB adaptation end ----")
        r_aw, g_aw, b_aw = rgb_aw
        logger.debug("RGB_AW: {}".format(rgb_aw))

        # ---------------------------
        # Opponent Color Dimensions
        # ---------------------------

        # achromatic_cone_signal
        a_a = 2 * r_a + g_a + (1 / 20) * b_a - 3.05 + 1
        logger.debug("A_A: {}".format(a_a))
        a_aw = 2 * r_aw + g_aw + (1 / 20) * b_aw - 3.05 + 1
        logger.debug("A_AW: {}".format(a_aw))

        c1 = r_a - g_a
        logger.debug("C1: {}".format(c1))
        c2 = g_a - b_a
        logger.debug("C2: {}".format(c2))
        c3 = b_a - r_a
        logger.debug("C3: {}".format(c3))

        c1_w = r_aw - g_aw
        logger.debug("C1_W: {}".format(c1_w))
        c2_w = g_aw - b_aw
        logger.debug("C2_W: {}".format(c2_w))
        c3_w = b_aw - r_aw
        logger.debug("C3_W: {}".format(c3_w))

        # -----
        # Hue
        # -----
        self._hue_angle = (
            180 * numpy.arctan2(0.5 * (c2 - c3) / 4.5, c1 - (c2 / 11)) / numpy.pi
        ) % 360
        hue_angle_w = (
            180
            * numpy.arctan2(0.5 * (c2_w - c3_w) / 4.5, c1_w - (c2_w / 11))
            / numpy.pi
        ) % 360

        # -------------
        # Saturation
        # -------------
        e_s = self._calculate_eccentricity_factor(self.hue_angle)
        logger.debug("es: {}".format(e_s))
        e_s_w = self._calculate_eccentricity_factor(hue_angle_w)

        f_t = l_a / (l_a + 0.1)
        logger.debug("F_t: {}".format(f_t))
        m_yb = 100 * (0.5 * (c2 - c3) / 4.5) * (e_s * (10 / 13) * n_c * n_cb * f_t)
        logger.debug("m_yb: {}".format(m_yb))
        m_rg = 100 * (c1 - (c2 / 11)) * (e_s * (10 / 13) * n_c * n_cb)
        logger.debug("m_rg: {}".format(m_rg))
        m = ((m_rg ** 2) + (m_yb ** 2)) ** 0.5
        logger.debug("m: {}".format(m))

        self._saturation = 50 * m / rgb_a.sum(axis=0)

        m_yb_w = (
            100 * (0.5 * (c2_w - c3_w) / 4.5) * (e_s_w * (10 / 13) * n_c * n_cb * f_t)
        )
        m_rg_w = 100 * (c1_w - (c2_w / 11)) * (e_s_w * (10 / 13) * n_c * n_cb)
        m_w = ((m_rg_w ** 2) + (m_yb_w ** 2)) ** 0.5

        # ------------
        # Brightness
        # ------------
        logger.debug("--- Stimulus achromatic signal START ----")
        a = self._calculate_achromatic_signal(l_as, s, s_w, n_bb, a_a)
        logger.debug("--- Stimulus achromatic signal END ----")
        logger.debug("A: {}".format(a))

        logger.debug("--- White achromatic signal START ----")
        a_w = self._calculate_achromatic_signal(l_as, s_w, s_w, n_bb, a_aw)
        logger.debug("--- White achromatic signal END ----")
        logger.debug("A_w: {}".format(a_w))

        n1 = ((7 * a_w) ** 0.5) / (5.33 * n_b ** 0.13)
        n2 = (7 * a_w * n_b ** 0.362) / 200
        logger.debug("N1: {}".format(n1))
        logger.debug("N2: {}".format(n2))

        self._brightness = ((7 * (a + (m / 100))) ** 0.6) * n1 - n2
        brightness_w = ((7 * (a_w + (m_w / 100))) ** 0.6) * n1 - n2
        logger.debug("Q: {}".format(self.brightness))
        logger.debug("Q_W: {}".format(brightness_w))

        # ----------
        # Lightness
        # ----------
        z = 1 + (y_b / y_w) ** 0.5
        logger.debug("z: {}".format(z))
        self._lightness = 100 * (self.brightness / brightness_w) ** z

        # -------
        # Chroma
        # -------
        self._chroma = (
            2.44
            * (self.saturation ** 0.69)
            * ((self.brightness / brightness_w) ** (y_b / y_w))
            * (1.64 - 0.29 ** (y_b / y_w))
        )

        # -------------
        # Colorfulness
        # -------------
        self._colorfulness = (f_l ** 0.15) * self.chroma

    xyz_to_rgb_m = numpy.array(
        [[0.38971, 0.68898, -0.07868], [-0.22981, 1.18340, 0.04641], [0, 0, 1]]
    )

    @classmethod
    def xyz_to_rgb(cls, xyz):
        return cls.xyz_to_rgb_m.dot(xyz)

    def _adaptation(
        self,
        f_l,
        l_a,
        xyz,
        xyz_w,
        xyz_b,
        xyz_p=None,
        p=None,
        helson_judd=False,
        discount_illuminant=True,
    ):
        """
        :param f_l: Luminance adaptation factor
        :param l_a: Adapting luminance
        :param xyz: Stimulus color in XYZ
        :param xyz_w: Reference white color in XYZ
        :param xyz_b: Background color in XYZ
        :param xyz_p: Proxima field color in XYZ
        :param p: Simultaneous contrast/assimilation parameter.
        """
        rgb = self.xyz_to_rgb(xyz)
        logger.debug("RGB: {}".format(rgb))
        rgb_w = self.xyz_to_rgb(xyz_w)
        logger.debug("RGB_W: {}".format(rgb_w))
        y_w = xyz_w[1]
        y_b = xyz_b[1]

        h_rgb = 3 * rgb_w / (rgb_w.sum())
        logger.debug("H_RGB: {}".format(h_rgb))

        # Chromatic adaptation factors
        if not discount_illuminant:
            f_rgb = (1 + (l_a ** (1 / 3)) + h_rgb) / (
                1 + (l_a ** (1 / 3)) + (1 / h_rgb)
            )
        else:
            f_rgb = numpy.ones(numpy.shape(h_rgb))
        logger.debug("F_RGB: {}".format(f_rgb))

        # Adaptation factor
        if helson_judd:
            d_rgb = self._f_n((y_b / y_w) * f_l * f_rgb[1]) - self._f_n(
                (y_b / y_w) * f_l * f_rgb
            )
            assert d_rgb[1] == 0
        else:
            d_rgb = numpy.zeros(numpy.shape(f_rgb))
        logger.debug("D_RGB: {}".format(d_rgb))

        # Cone bleaching factors
        rgb_b = (10 ** 7) / ((10 ** 7) + 5 * l_a * (rgb_w / 100))
        logger.debug("B_RGB: {}".format(rgb_b))

        if xyz_p is not None and p is not None:
            logger.debug("Account for simultaneous chromatic contrast")
            rgb_p = self.xyz_to_rgb(xyz_p)
            rgb_w = self.adjust_white_for_scc(rgb_p, rgb_b, rgb_w, p)

        # Adapt rgb using modified
        rgb_a = 1 + rgb_b * (self._f_n(f_l * f_rgb * rgb / rgb_w) + d_rgb)
        logger.debug("RGB_A: {}".format(rgb_a))

        return rgb_a

    @classmethod
    def adjust_white_for_scc(cls, rgb_p, rgb_b, rgb_w, p):
        """
        Adjust the white point for simultaneous chromatic contrast.

        :param rgb_p: Cone signals of proxima field.
        :param rgb_b: Cone signals of background.
        :param rgb_w: Cone signals of reference white.
        :param p: Simultaneous contrast/assimilation parameter.
        :return: Adjusted cone signals for reference white.
        """
        p_rgb = rgb_p / rgb_b
        rgb_w = (
            rgb_w
            * (((1 - p) * p_rgb + (1 + p) / p_rgb) ** 0.5)
            / (((1 + p) * p_rgb + (1 - p) / p_rgb) ** 0.5)
        )
        return rgb_w

    @staticmethod
    def _get_cct(x, y, z):
        """
        Reference
        Hernandez-Andres, J., Lee, R. L., & Romero, J. (1999).
        Calculating correlated color temperatures across the entire gamut of daylight
        and skylight chromaticities.
        Applied Optics, 38(27), 5703-5709.
        """
        x_e = 0.3320
        y_e = 0.1858

        n = ((x / (x + z + z)) - x_e) / ((y / (x + z + z)) - y_e)

        a_0 = -949.86315
        a_1 = 6253.80338
        a_2 = 28.70599
        a_3 = 0.00004

        t_1 = 0.92159
        t_2 = 0.20039
        t_3 = 0.07125

        cct = (
            a_0
            + a_1 * numpy.exp(-n / t_1)
            + a_2 * numpy.exp(-n / t_2)
            + a_3 * numpy.exp(-n / t_3)
        )
        return cct

    @staticmethod
    def calculate_scotopic_luminance(photopic_luminance, color_temperature):
        return 2.26 * photopic_luminance * ((color_temperature / 4000) - 0.4) ** (1 / 3)

    @classmethod
    def _calculate_achromatic_signal(cls, l_as, s, s_w, n_bb, a_a):

        j = 0.00001 / ((5 * l_as / 2.26) + 0.00001)
        logger.debug("j: {}".format(j))

        f_ls = 3800 * (j ** 2) * (5 * l_as / 2.26)
        f_ls += 0.2 * ((1 - (j ** 2)) ** 0.4) * ((5 * l_as / 2.26) ** (1 / 6))
        logger.debug("F_LS: {}".format(f_ls))

        b_s = 0.5 / (1 + 0.3 * ((5 * l_as / 2.26) * (s / s_w)) ** 0.3)
        b_s += 0.5 / (1 + 5 * (5 * l_as / 2.26))
        logger.debug("B_S: {}".format(b_s))

        a_s = (cls._f_n(f_ls * s / s_w) * 3.05 * b_s) + 0.3
        logger.debug("A_S: {}".format(a_s))

        return n_bb * (a_a - 1 + a_s - 0.3 + numpy.sqrt((1 + (0.3 ** 2))))

    @staticmethod
    def _f_n(i):
        """
        Nonlinear response function.
        """
        return 40 * ((i ** 0.73) / (i ** 0.73 + 2))

    @staticmethod
    def _calculate_eccentricity_factor(hue_angle):
        h = numpy.array([20.14, 90, 164.25, 237.53])
        e = numpy.array([0.8, 0.7, 1.0, 1.2])

        out = numpy.interp(hue_angle, h, e)
        out = numpy.where(hue_angle < 20.14, 0.856 - (hue_angle / 20.14) * 0.056, out)
        out = numpy.where(
            hue_angle > 237.53, 0.856 + 0.344 * (360 - hue_angle) / (360 - 237.53), out
        )

        return out


class RLAB(object):
    """
    **References**

    *   Fairchild, M. D. (1996). Refinement of the RLAB color space.
        *Color Research & Application*, 21(5), 338-346.
    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.

    """

    @property
    def hue_angle(self):
        """
        Predicted hue angle :math:`h^R`.
        """
        return self._hue_angle

    @property
    def chroma(self):
        """
        Predicted chroma :math:`C^R`.
        """
        return self._chroma

    @property
    def saturation(self):
        """
        Predicted saturation :math:`s^R`.
        """
        return self._saturation

    @property
    def lightness(self):
        """
        Predicted colorfulness :math:`L^R`.
        """
        return self._lightness

    @property
    def a(self):
        """
        Predicted red-green chromatic response :math:`a^R`.
        """
        return self._a

    @property
    def b(self):
        """
        Predicted yellow-blue chromatic response :math:`b^R`.
        """
        return self._b

    R = numpy.array([[1.9569, -1.1882, 0.2313], [0.3612, 0.6388, 0], [0, 0, 1]])

    def __init__(self, x, y, z, x_n, y_n, z_n, y_n_abs, sigma, d):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_n: X value of reference white :math:`X_n`.
        :param y_n: Y value of reference white :math:`Y_n`.
        :param z_n: Z value of reference white :math:`Z_n`.
        :param y_n_abs: Absolute luminance :math:`Y_n` of a white object in cd/m^2.
        :param sigma: Relative luminance parameter :math:`\\sigma`. For average
                      surround set :math:`\\sigma=1/2.3`, for dim surround
                      :math:`\\sigma=1/2.9` and for dark surround
                      :math:`\\sigma=1/3.5`.
        :param d: Degree of adaptation :math:`D`.
        """
        xyz = numpy.array([x, y, z])
        xyz_n = numpy.array([x_n, y_n, z_n])

        lms = Hunt.xyz_to_rgb(xyz)
        lms_n = Hunt.xyz_to_rgb(xyz_n)
        logger.debug("LMS: {}".format(lms))
        logger.debug("LMS_n: {}".format(lms_n))

        lms_e = (3 * lms_n) / (lms_n[0] + lms_n[1] + lms_n[2])
        lms_p = (1 + (y_n_abs ** (1 / 3)) + lms_e) / (
            1 + (y_n_abs ** (1 / 3)) + (1 / lms_e)
        )
        logger.debug("LMS_e: {}".format(lms_e))
        logger.debug("LMS_p: {}".format(lms_p))

        lms_a = (lms_p + d * (1 - lms_p)) / lms_n
        logger.debug("LMS_a: {}".format(lms_a))

        # If we want to allow arrays as input we need special handling here.
        if len(numpy.shape(x)) == 0:
            # Okay so just a number, we can do things by the book.
            a = numpy.diag(lms_a)
            logger.debug("A: {}".format(a))
            xyz_ref = self.R.dot(a).dot(Hunt.xyz_to_rgb_m).dot(xyz)
        else:
            # So we have an array. Since constructing huge multidimensional
            # arrays might not bee the best idea, we will handle each input
            # dimension separately. First figure out how many values we have to
            # deal with.
            input_dim = len(x)
            # No create the output array that we will fill layer by layer
            xyz_ref = numpy.zeros((3, input_dim))
            for layer in range(input_dim):
                a = numpy.diag(lms_a[..., layer])
                logger.debug("A layer {}: {}".format(layer, a))
                xyz_ref[..., layer] = (
                    self.R.dot(a).dot(Hunt.xyz_to_rgb_m).dot(xyz[..., layer])
                )

        logger.debug("XYZ_ref: {}".format(xyz_ref))
        x_ref, y_ref, z_ref = xyz_ref

        # Lightness
        self._lightness = 100 * (y_ref ** sigma)
        logger.debug("lightness: {}".format(self.lightness))

        # Opponent Color Dimensions
        self._a = 430 * ((x_ref ** sigma) - (y_ref ** sigma))
        self._b = 170 * ((y_ref ** sigma) - (z_ref ** sigma))
        logger.debug("a: {}".format(self._a))
        logger.debug("b: {}".format(self._b))

        # Hue
        self._hue_angle = (
            360 * numpy.arctan2(self._b, self._a) / (2 * numpy.pi) + 360
        ) % 360

        # Chroma
        self._chroma = numpy.sqrt((self._a ** 2) + (self._b ** 2))

        # Saturation
        self._saturation = self.chroma / self.lightness


class ATD95(object):
    """
    **References**


    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.
    *   Guth, S. L. (1995, April). Further applications of the ATD model for color
        vision. In *IS&T/SPIE's Symposium on Electronic Imaging: Science &
        Technology* (pp. 12-26). International Society for Optics and Photonics.

    """

    @property
    def hue(self):
        """
        Predicted hue :math:`H`.
        """
        return self._hue

    @property
    def brightness(self):
        """
        Predicted brightness :math:`Br`.
        """
        return self._brightness

    @property
    def saturation(self):
        """
        Predicted saturation :math:`C`.
        """
        return self._saturation

    def __init__(self, x, y, z, x_0, y_0, z_0, y_0_abs, k_1, k_2, sigma=300):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_0: X value of reference white :math:`X_0`.
        :param y_0: Y value of reference white :math:`Y_0`.
        :param z_0: Z value of reference white :math:`Z_0`.
        :param y_0_abs: Absolute adapting luminance :math:`Y_0` in cd/m^2.
        :param k_1: :math:`k_1`
        :param k_2: :math:`k_2`
        :param sigma: :math:`\\sigma`
        """
        xyz = self._scale_to_luminance(numpy.array([x, y, z]), y_0_abs)
        xyz_0 = self._scale_to_luminance(numpy.array([x_0, y_0, z_0]), y_0_abs)
        logger.debug("Scaled XYZ: {}".format(xyz))
        logger.debug("Scaled XYZ_0: {}".format(xyz))

        # Adaptation Model
        lms = self._xyz_to_lms(xyz)
        logger.debug("LMS: {}".format(lms))

        xyz_a = k_1 * xyz + k_2 * xyz_0
        logger.debug("XYZ_a: {}".format(xyz_a))

        lms_a = self._xyz_to_lms(xyz_a)
        logger.debug("LMS_a: {}".format(lms_a))

        l_g, m_g, s_g = lms * (sigma / (sigma + lms_a))

        # Opponent Color Dimensions
        a_1i = 3.57 * l_g + 2.64 * m_g
        t_1i = 7.18 * l_g - 6.21 * m_g
        d_1i = -0.7 * l_g + 0.085 * m_g + s_g
        a_2i = 0.09 * a_1i
        t_2i = 0.43 * t_1i + 0.76 * d_1i
        d_2i = d_1i

        self._a_1 = self._calculate_final_response(a_1i)
        self._t_1 = self._calculate_final_response(t_1i)
        self._d_1 = self._calculate_final_response(d_1i)
        self._a_2 = self._calculate_final_response(a_2i)
        self._t_2 = self._calculate_final_response(t_2i)
        self._d_2 = self._calculate_final_response(d_2i)

        # Perceptual Correlates
        self._brightness = (self._a_1 ** 2 + self._t_1 ** 2 + self._d_1 ** 2) ** 0.5
        self._saturation = (self._t_2 ** 2 + self._d_2 ** 2) ** 0.5 / self._a_2
        self._hue = self._t_2 / self._d_2

    @staticmethod
    def _calculate_final_response(value):
        return value / (200 + abs(value))

    @staticmethod
    def _scale_to_luminance(xyz, absolute_adapting_luminance):
        return 18 * (absolute_adapting_luminance * xyz / 100) ** 0.8

    @staticmethod
    def _xyz_to_lms(xyz):
        x, y, z = xyz
        L = ((0.66 * (0.2435 * x + 0.8524 * y - 0.0516 * z)) ** 0.7) + 0.024
        M = ((-0.3954 * x + 1.1642 * y + 0.0837 * z) ** 0.7) + 0.036
        S = ((0.43 * (0.04 * y + 0.6225 * z)) ** 0.7) + 0.31
        return numpy.array([L, M, S])


class LLAB(object):
    """
    **References**

    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.
    *   Luo, M. R., & Morovic, J. (1996, September). Two unsolved issues in colour
        management-colour appearance and gamut mapping. In *5th International
        Conference on High Technology* (pp. 136-147).
    *   Luo, M. R., Lo, M. C., & Kuo, W. G. (1996). The LLAB (l: c) colour model.
        *Color Research & Application*, 21(6), 412-429.
    """

    @property
    def hue_angle(self):
        """
        Predicted hue angle :math:`h_L`.
        """
        return self._hue_angle

    @property
    def chroma(self):
        """
        Predicted chroma :math:`Ch_L`.
        """
        return self._chroma

    @property
    def saturation(self):
        """
        Predicted saturation :math:`s_L`.
        """
        return self._saturation

    @property
    def lightness(self):
        """
        Predicted colorfulness :math:`L_L`.
        """
        return self._lightness

    @property
    def a_l(self):
        """
        Predicted red-green chromatic response :math:`A_L`.
        """
        return self._a_l

    @property
    def b_l(self):
        """
        Predicted yellow-blue chromatic response :math:`B_L`.
        """
        return self._b_l

    def __init__(self, x, y, z, x_0, y_0, z_0, y_b, f_s, f_l, f_c, l, d=1):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_0: X value of reference white :math:`X_0`.
        :param y_0: Y value of reference white :math:`Y_0`.
        :param z_0: Z value of reference white :math:`Z_0`.
        :param y_b: Luminance factor of the background :math:`Y_b` in cd/m^2.
        :param f_s: Surround induction factor :math:`F_S`.
        :param f_l: Lightness induction factor :math:`F_L`.
        :param f_c: Chroma induction factor :math:`F_C`.
        :param l: Absolute luminance of reference white :math:`L` in cd/m^2.
        :param d: Discounting-the-Illuminant factor :math:`D`.
        """
        xyz = numpy.array([x, y, z])
        logger.debug("XYZ: {}".format([x, y, z]))
        xyz_0 = numpy.array([x_0, y_0, z_0])

        r, g, b = self.xyz_to_rgb(xyz)
        logger.debug("RGB: {}".format([r, g, b]))
        r_0, g_0, b_0 = self.xyz_to_rgb(xyz_0)
        logger.debug("RGB_0: {}".format([r_0, g_0, b_0]))

        xyz_0r = numpy.array([95.05, 100, 108.88])
        r_0r, g_0r, b_0r = self.xyz_to_rgb(xyz_0r)
        logger.debug("RGB_0r: {}".format([r_0r, g_0r, b_0r]))

        beta = (b_0 / b_0r) ** 0.0834
        logger.debug("beta: {}".format(beta))
        r_r = (d * (r_0r / r_0) + 1 - d) * r
        g_r = (d * (g_0r / g_0) + 1 - d) * g
        b_r = (d * (b_0r / (b_0 ** beta)) + 1 - d) * (abs(b) ** beta)
        logger.debug("RGB_r: {}".format([r_r, g_r, b_r]))

        rgb_r = numpy.array([r_r, g_r, b_r])

        # m_inv = numpy.linalg.inv(self.xyz_to_rgb_m)
        m_inv = numpy.array(
            [[0.987, -0.1471, 0.16], [0.4323, 0.5184, 0.0493], [-0.0085, 0.04, 0.9685]]
        )
        x_r, y_r, z_r = m_inv.dot(rgb_r * y)
        logger.debug("XYZ_r: {}".format([x_r, y_r, z_r]))

        # Opponent Color Dimension
        def f(w):
            return numpy.where(
                w > 0.008856,
                w ** (1 / f_s),
                (((0.008856 ** (1 / f_s)) - (16 / 116)) / 0.008856) * w + (16 / 116),
            )

        # lightness_contrast_exponent
        z = 1 + f_l * ((y_b / 100) ** 0.5)
        logger.debug("z: {}".format(z))

        self._lightness = 116 * (f(y_r / 100) ** z) - 16
        a = 500 * (f(x_r / 95.05) - f(y_r / 100))
        b = 200 * (f(y_r / 100) - f(z_r / 108.88))
        logger.debug("A: {}".format(a))
        logger.debug("B: {}".format(b))

        logger.debug("f(Xr): {}".format(f(x_r / 95.05)))
        logger.debug("f(Yr): {}".format(f(y_r / 100)))
        logger.debug("f(Zr): {}".format(f(z_r / 108.88)))

        # Perceptual Correlates
        c = (a ** 2 + b ** 2) ** 0.5
        self._chroma = 25 * numpy.log(1 + 0.05 * c)

        s_c = 1 + 0.47 * numpy.log10(l) - 0.057 * numpy.log10(l) ** 2
        s_m = 0.7 + 0.02 * self._lightness - 0.0002 * self._lightness ** 2
        c_l = self._chroma * s_m * s_c * f_c

        self._saturation = self._chroma / self._lightness

        hue_angle_rad = numpy.arctan2(b, a)
        self._hue_angle = hue_angle_rad * 360 / (2 * numpy.pi) % 360

        self._a_l = c_l * numpy.cos(hue_angle_rad)
        self._b_l = c_l * numpy.sin(hue_angle_rad)

    xyz_to_rgb_m = numpy.array(
        [
            [0.8951, 0.2664, -0.1614],
            [-0.7502, 1.7135, 0.0367],
            [0.0389, -0.0685, 1.0296],
        ]
    )

    @classmethod
    def xyz_to_rgb(cls, xyz):
        return cls.xyz_to_rgb_m.dot(xyz / xyz[1])


class CIECAM02(object):
    """
    **References**

    *   CIE TC 8-01 (2004). A Color appearance model for color management systems.
        Publication 159. Vienna: CIE Central Bureau. ISBN 3-901906-29-0.
    *   Fairchild, M. D. (2013). *Color appearance models*, 3rd Ed. John Wiley & Sons.
    """

    @property
    def hue_angle(self):
        """
        Predicted hue angle :math:`h`.
        """
        return self._h

    @property
    def chroma(self):
        """
        Predicted chroma :math:`C`.
        """
        return self._chroma

    @property
    def saturation(self):
        """
        Predicted saturation :math:`s_L`.
        """
        return self._saturation

    @property
    def lightness(self):
        """
        Predicted colorfulness :math:`J`.
        """
        return self._lightness

    @property
    def brightness(self):
        """
        Predicted colorfulness :math:`Q`.
        """
        return self._brightness

    @property
    def colorfulness(self):
        """
        Predicted colorfulness :math:`M`.
        """
        return self._colorfulness

    @property
    def a(self):
        """
        Predicted red-green chromatic response :math:`a`.
        """
        return self._a

    @property
    def b(self):
        """
        Predicted yellow-blue chromatic response :math:`b`.
        """
        return self._b

    M_CAT02 = numpy.array(
        [[0.7328, 0.4296, -0.1624], [-0.7036, 1.6975, 0.0061], [0.0030, 0.0136, 0.9834]]
    )

    M_CAT02_inv = numpy.linalg.inv(M_CAT02)

    M_HPE = numpy.array(
        [[0.38971, 0.68898, -0.07868], [-0.22981, 1.18340, 0.04641], [0, 0, 1]]
    )

    def __init__(self, x, y, z, x_w, y_w, z_w, y_b, l_a, c, n_c, f, d=False):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_w: X value of reference white :math:`X_W`.
        :param y_w: Y value of reference white :math:`Y_W`.
        :param z_w: Z value of reference white :math:`Z_W`.
        :param y_b: Background relative luminance :math:`Y_b`.
        :param l_a: Adapting luminance :math:`L_A` in cd/m^2.
        :param c: Exponential nonlinearity :math:`c`. (Average/Dim/Dark)
                  (0.69/0.59/0.525).
        :param n_c: Chromatic induction factor :math:`N_c`. (Average/Dim/Dark)
                    (1.0,0.9,0.8).
        :param f: Maximum degree of adaptation :math:`F`. (Average/Dim/Dark)
                  (1.0/0.9/0.8).
        :param d: Discount-the-Illuminant factor :math:`D`.
        """

        xyz = numpy.array([x, y, z])
        xyz_w = numpy.array([x_w, y_w, z_w])

        # Determine the degree of adaptation
        if not d:
            d = self._compute_degree_of_adaptation(f, l_a)
        else:
            d = 1
        logger.debug("D: {}".format(d))

        # Compute viewing condition dependant components
        k = 1 / (5 * l_a + 1)
        logger.debug("k: {}".format(k))

        f_l = 0.2 * (k ** 4) * 5 * l_a + 0.1 * (1 - k ** 4) ** 2 * (5 * l_a) ** (1 / 3)
        logger.debug("F_L: {}".format(f_l))
        n = y_b / y_w
        logger.debug("n: {}".format(n))
        self.n_bb = self.n_cb = 0.725 * n ** -0.2
        z = 1.48 + numpy.sqrt(n)
        logger.debug("z: {}".format(z))

        rgb_a, rgb_aw = self._compute_adaptation(xyz, xyz_w, f_l, d)
        logger.debug("RGB'a: {}".format(rgb_a))
        logger.debug("RGB'aw: {}".format(rgb_aw))

        r_a, g_a, b_a = rgb_a
        r_aw, g_aw, b_aw = rgb_aw

        # Opponent Color Dimensions
        self._a = r_a - 12 * g_a / 11 + b_a / 11
        self._b = (1 / 9) * (r_a + g_a - 2 * b_a)

        # Hue
        self._h = 360 * numpy.arctan2(self._b, self._a) / (2 * numpy.pi)
        e_t = (1 / 4) * (numpy.cos(2 + self._h * numpy.pi / 180) + 3.8)

        # Lightness
        a = self._compute_achromatic_response(r_a, g_a, b_a, self.n_bb)
        logger.debug("A: {}".format(a))
        a_w = self._compute_achromatic_response(r_aw, g_aw, b_aw, self.n_bb)
        logger.debug("A_W: {}".format(a_w))
        self._lightness = 100 * (a / a_w) ** (c * z)  # 16.24

        # Brightness
        # self._brightness = self.compute_brightness(self.lightness, surround, a_w, f_l)
        self._brightness = (
            (4 / c) * numpy.sqrt(self._lightness / 100) * (a_w + 4) * f_l ** 0.25
        )

        # Chroma
        # self.chroma = self.compute_chroma(rgb_a, self.lightness, surround,
        # self.N_cb, e_t, self.a, self.b, n)
        t = (
            (50000 / 13)
            * n_c
            * self.n_cb
            * e_t
            * numpy.sqrt((self._a ** 2) + (self._b ** 2))
        ) / (rgb_a[0] + rgb_a[1] + (21 / 20) * rgb_a[2])
        self._chroma = (
            (t ** 0.9)
            * numpy.sqrt(self._lightness / 100)
            * ((1.64 - 0.29 ** n) ** 0.73)
        )

        # Colorfulness
        self._colorfulness = self.chroma * f_l ** 0.25

        # Saturation
        self._saturation = 100 * numpy.sqrt(self._colorfulness / self._brightness)

        # Cartesian coordinates
        self.a_c, self.b_c = self._compute_cartesian_coordinates(self.chroma, self._h)
        self.a_m, self.b_m = self._compute_cartesian_coordinates(
            self._colorfulness, self._h
        )
        self.a_s, self.b_s = self._compute_cartesian_coordinates(
            self.saturation, self._h
        )

    @classmethod
    def _compute_adaptation(cls, xyz, xyz_w, f_l, d):
        # Transform input colors to cone responses
        rgb = cls._xyz_to_rgb(xyz)
        logger.debug("RGB: {}".format(rgb))
        rgb_w = cls._xyz_to_rgb(xyz_w)
        logger.debug("RGB_W: {}".format(rgb_w))

        # Compute adapted tristimulus-responses
        rgb_c = cls._white_adaption(rgb, rgb_w, d)
        logger.debug("RGB_C: {}".format(rgb_c))
        rgb_cw = cls._white_adaption(rgb_w, rgb_w, d)
        logger.debug("RGB_CW: {}".format(rgb_cw))

        # Convert adapted tristimulus-responses to Hunt-Pointer-Estevez fundamentals
        rgb_p = cls._compute_hunt_pointer_estevez_fundamentals(rgb_c)
        logger.debug("RGB': {}".format(rgb_p))
        rgb_wp = cls._compute_hunt_pointer_estevez_fundamentals(rgb_cw)
        logger.debug("RGB'_W: {}".format(rgb_wp))

        # Compute post-adaptation non-linearities
        rgb_ap = cls._compute_nonlinearities(f_l, rgb_p)
        rgb_awp = cls._compute_nonlinearities(f_l, rgb_wp)

        return rgb_ap, rgb_awp

    @staticmethod
    def _xyz_to_rgb(xyz):
        return numpy.dot(CIECAM02.M_CAT02, xyz)

    @staticmethod
    def _rgb_to_xyz(rgb):
        return numpy.dot(CIECAM02.M_CAT02_inv, rgb)

    @staticmethod
    def _white_adaption(rgb, rgb_w, d=1):
        return ((100 * d / rgb_w) + (1 - d)) * rgb

    @staticmethod
    def _compute_degree_of_adaptation(surround_conditions, adapting_luminance):
        return surround_conditions * (
            1 - (1 / 3.6) * numpy.exp((-adapting_luminance - 42) / 92)
        )

    @staticmethod
    def _compute_hunt_pointer_estevez_fundamentals(rgb):
        return numpy.dot(numpy.dot(CIECAM02.M_HPE, CIECAM02.M_CAT02_inv), rgb)

    @staticmethod
    def _compute_nonlinearities(f_l, rgb):
        return 0.1 + (400 * (f_l * rgb / 100) ** 0.42) / (
            27.13 + (f_l * rgb / 100) ** 0.42
        )

    @staticmethod
    def _compute_achromatic_response(r, g, b, n_bb):
        return (2 * r + g + (1 / 20) * b - 0.305) * n_bb

    @staticmethod
    def _compute_cartesian_coordinates(value, hue):
        a = value * numpy.cos(hue * numpy.pi / 180)  # 16.30
        b = value * numpy.sin(hue * numpy.pi / 180)  # 16.31
        return a, b


class CIECAM02m1(CIECAM02):
    """
    **References**

    *   Wu, R. C., & Wardman, R. H. (2007). Proposed modification to the CIECAM02
        colour appearance model to include the simultaneous contrast effects.
        *Color Research & Application*, 32(2), 121-129.
    """

    def __init__(
        self, x, y, z, x_w, y_w, z_w, x_b, y_b, z_b, l_a, c, n_c, f, p, d=False
    ):
        """
        :param x: X value of test sample :math:`X`.
        :param y: Y value of test sample :math:`Y`.
        :param z: Z value of test sample :math:`Z`.
        :param x_w: X value of reference white :math:`X_W`.
        :param y_w: Y value of reference white :math:`Y_W`.
        :param z_w: Z value of reference white :math:`Z_W`.
        :param x_b: X value of background :math:`X_b`.
        :param y_b: Y value of background :math:`Y_b`.
        :param z_b: Z value of background :math:`Z_b`.
        :param l_a: Adapting luminance :math:`L_A` in cd/m^2.
        :param c: Exponential nonlinearity :math:`c`. (Average/Dim/Dark)
                  (0.69/0.59/5.25).
        :param n_c: Chromatic induction factor :math:`N_c`. (Average/Dim/Dark)
                    (1.0,0.9,0.8).
        :param f: Maximum degree of adaptation :math:`F`. (Average/Dim/Dark)
                  (1.0/0.9/0.8).
        :param p: Simultaneous contrast/assimilation parameter.
        :param d: Discount-the-Illuminant factor :math:`D`.
        """

        self._p = p
        self._xyz_b = numpy.array([x_b, y_b, z_b])

        super(CIECAM02m1, self).__init__(x, y, z, x_w, y_w, z_w, y_b, l_a, c, n_c, f, d)

    def _compute_adaptation(self, xyz, xyz_w, f_l, d):
        """
        Modified adaptation procedure incorporating simultaneous chromatic contrast
        from Hunt model.

        :param xyz: Stimulus XYZ.
        :param xyz_w: Reference white XYZ.
        :param f_l: Luminance adaptation factor
        :param d: Degree of adaptation.
        :return: Tuple of adapted rgb and rgb_w arrays.
        """
        # Transform input colors to cone responses
        rgb = self._xyz_to_rgb(xyz)
        logger.debug("RGB: {}".format(rgb))

        rgb_b = self._xyz_to_rgb(self._xyz_b)
        rgb_w = self._xyz_to_rgb(xyz_w)
        rgb_w = Hunt.adjust_white_for_scc(rgb, rgb_b, rgb_w, self._p)
        logger.debug("RGB_W: {}".format(rgb_w))

        # Compute adapted tristimulus-responses
        rgb_c = self._white_adaption(rgb, rgb_w, d)
        logger.debug("RGB_C: {}".format(rgb_c))
        rgb_cw = self._white_adaption(rgb_w, rgb_w, d)
        logger.debug("RGB_CW: {}".format(rgb_cw))

        # Convert adapted tristimulus-responses to Hunt-Pointer-Estevez fundamentals
        rgb_p = self._compute_hunt_pointer_estevez_fundamentals(rgb_c)
        logger.debug("RGB': {}".format(rgb_p))
        rgb_wp = self._compute_hunt_pointer_estevez_fundamentals(rgb_cw)
        logger.debug("RGB'_W: {}".format(rgb_wp))

        # Compute post-adaptation non-linearities
        rgb_ap = self._compute_nonlinearities(f_l, rgb_p)
        rgb_awp = self._compute_nonlinearities(f_l, rgb_wp)

        return rgb_ap, rgb_awp



from kivy.config import Config
Config.set('graphics', 'width', '400')
Config.set('graphics', 'height', '400')
from kivy.app import App
from kivy.graphics import Ellipse, Rectangle, Color
from kivy.uix.widget import Widget
from colorir import config, KIVY_COLOR_FORMAT, Grad, HSV

# Define the default color format to work with kivy
config.DEFAULT_COLOR_FORMAT = KIVY_COLOR_FORMAT

# You can play around with these variables to change the appearance of the color wheel
ELLIPSE_STEPS = 150
ELLIPSE_SEGMENTS = 50


class WheelScreen(Widget):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        iter_angle = 360 / ELLIPSE_STEPS
        with self.canvas:
            Rectangle(pos=(0, 0), size=(400, 400))  # Draw background

            for j in range(ELLIPSE_SEGMENTS + 1):
                s = 1 - j / ELLIPSE_SEGMENTS
                # Sample colors throughout the hue range for this saturation
                outer_colors = Grad(
                    [HSV(0, s, 1), HSV(355, s, 1)],
                    color_sys=HSV
                ).n_colors(ELLIPSE_STEPS)
                for i, color in enumerate(outer_colors):
                    Color(*color)
                    width = 400 * (ELLIPSE_SEGMENTS - j) / ELLIPSE_SEGMENTS
                    height = 400 * (ELLIPSE_SEGMENTS - j) / ELLIPSE_SEGMENTS
                    Ellipse(
                        pos=(200 - width / 2, 200 - height / 2),
                        size=(width, height),
                        angle_start=i * iter_angle,
                        angle_end=(i + 1) * iter_angle
                    )


class WheelApp(App):
    def build(self):
        return WheelScreen()


app = WheelApp()
app.run()



import numpy as np
import matplotlib.pyplot as plt
from colorir import *

fig, axes = plt.subplots(2, 2)
imdata = (np.outer(np.linspace(-1, 1, 100), np.linspace(-1, 1, 100)) + 1) / 2
cs = Palette.load()  # Load all colors available

# Discrete colormap from palette
pal = StackPalette.load("carnival").resize(8)  # Load the palette and resize it to get 8 color categories
ys = [np.random.random(10) + 1 for i in range(len(pal))]
axes[0, 0].stackplot(np.arange(10), *ys, colors=pal)
axes[0, 0].set_title("Stacked data")

# Discrete PuBu & GnBu
pubu = StackPalette.load("pubu")
gnbu = StackPalette.load("gnbu")
# Fuse palettes with the & operator
pal = pubu & gnbu[::-1]
# Make a discrete gradient from the palette
grad = Grad(pal.resize(13), discrete=True)
im1 = axes[0, 1].imshow(imdata, cmap=grad.to_cmap())
axes[0, 1].set_title("Discrete PuBu & GnBu")

# Divergent purple and yellow gradient
grad = PolarGrad([cs.cyan, cs.eggyolk, cs.magenta])
im2 = axes[1, 0].imshow(imdata, cmap=grad.to_cmap())
axes[1, 0].set_title("Divergent cyan magenta")

# Non-linear gradient
# Place the white closer to the end of the gradient instead of in the middle by setting 'color_coords'
grad = PolarGrad([cs.darkred, cs.white, cs.darkblue], color_coords=[0, 2 / 3, 1])
im3 = axes[1, 1].imshow(imdata, cmap=grad.to_cmap())
axes[1, 1].set_title("Non-linear gradient")

# Make stuff look good
axes[0, 0].set_aspect(1/2)
axes[0, 0].set_xlim([0, 9])
plt.colorbar(im1)
plt.colorbar(im2)
plt.colorbar(im3)

fig.tight_layout()
plt.show()



import doctest
import unittest
from colorir import *

config.REPR_STYLE = "traditional"


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(utils))
    return tests


class TestDistance(unittest.TestCase):
    def test_simple_dist(self):
        dist = utils.simplified_dist(RGB(1, 1, 1), RGB(0, 0, 0))
        self.assertAlmostEqual(dist, 765)


if __name__ == "__main__":
    unittest.main()


import doctest
import unittest
from pathlib import Path
from random import randint
import re

from colorir import *

config.REPR_STYLE = "traditional"
config.DEFAULT_PALETTES_DIR = str(Path(__file__).resolve().parent / "test_palettes")
# Clear test palette directory
for file in Path(config.DEFAULT_PALETTES_DIR).glob("*"):
    # Keep original test files, remove others
    if re.search(r"test\d", file.name) is None:
        file.unlink()


class TestPalette(unittest.TestCase):
    def test_and_op(self):
        pal = Palette(c1="ffffff") & Palette(c2="000000")
        self.assertEqual(pal, Palette(c1="ffffff", c2="000000"))

    def test_save_load(self):
        colors = {f"c{i}": "%02x%02x%02x" % (randint(0, 255), randint(0, 255), randint(0, 255))
                  for i in range(250)}
        pal = Palette(colors)
        pal.save(name="test_sl")
        pal2 = Palette.load("test_sl")
        self.assertEqual(pal, pal2)

    def test_load_warns(self):
        with self.assertWarns(Warning):
            Palette.load(palettes=["test1", "test2"], search_builtins=False)

    def test_color_manipulation(self):
        pal = Palette(red="ff0000", green="00ff00", blue="0000ff")
        manip_dict = {"red": HCLuv(1, 0.5, 1), "green": CIELab(0.2, 1, 1), "blue": HSL(0.5, 1, 1)}
        self.assertEqual(pal * manip_dict, Palette(red="#d05959", green="#003d00", blue="#00ff00"))


class TestStackPalette(unittest.TestCase):
    def test_and_op(self):
        spal = StackPalette(["ffffff"]) & StackPalette(["000000"])
        self.assertEqual(spal, StackPalette(["ffffff", "000000"]))

    def test_save_load(self):
        colors = ["%02x%02x%02x" % (randint(0, 255), randint(0, 255), randint(0, 255))
                  for _ in range(250)]
        spal = StackPalette(colors)
        spal.save(name="test_sl")
        spal2 = StackPalette.load("test_sl")
        self.assertEqual(spal, spal2)

    def test_complementary(self):
        red = "ff0000"
        spal = StackPalette.new_complementary(3, red)
        self.assertEqual(spal, StackPalette([red, "00ff00", "0000ff"]))

    def test_analogous(self):
        red = HSV(0, 1, 1)
        # Center, clockwise and counter-clockwise generation
        c_spal = StackPalette.new_analogous(3, color=red)
        cw_spal = StackPalette.new_analogous(3, color=red, start=1)
        ccw_spal = StackPalette.new_analogous(3, color=red, start=-1)
        self.assertEqual(c_spal, StackPalette([HSV(330, 1, 1), red, HSV(30, 1, 1)]))
        self.assertEqual(cw_spal, StackPalette([red, HSV(30, 1, 1), HSV(60, 1, 1)]))
        self.assertEqual(ccw_spal, StackPalette([HSV(300, 1, 1), HSV(330, 1, 1), red]))
        # Larger hue wheel sections
        wide_spal = StackPalette.new_analogous(3, sections=3, color=red)
        self.assertEqual(wide_spal, StackPalette([HSV(240, 1, 1), red, HSV(120, 1, 1)]))

    def test_color_manipulation(self):
        pal = StackPalette(["ff0000", "00ff00", "0000ff"])
        manip_dict = [HCLuv(1, 0.5, 1), CIELab(0.2, 1, 1), HSL(0.5, 1, 1)]
        self.assertEqual(pal * manip_dict, StackPalette(["#d05959", "#003d00", "#00ff00"]))


class TestOtherFunctions(unittest.TestCase):
    def test_find_palettes(self):
        all_pals = find_palettes(search_builtins=False)
        self.assertEqual(all_pals, ['test1', 'test2', 'test3'])
        pals = find_palettes(search_builtins=False, kind=Palette)
        self.assertEqual(pals, ['test1', 'test2'])
        spals = find_palettes(search_builtins=False, kind=StackPalette)
        self.assertEqual(spals, ['test3'])

    def test_delete_palette(self):
        n_pal = Palette(red="ff0000")
        n_pal.save("test_del")
        self.assertIn("test_del", find_palettes())
        delete_palette("test_del")
        self.assertNotIn("test_del", find_palettes())


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(palette))
    return tests


if __name__ == "__main__":
    unittest.main()


import csv
import numpy as np
import plotly.graph_objs as go
from pathlib import Path
from plotly.io import templates as plotly_templates
from plotly.subplots import make_subplots
from colorir import *

palettes = {name: Palette.load(name).to_stackpalette() for name in find_palettes(kind=Palette) if name != "css"}
for name in find_palettes(kind=StackPalette):
    dict_name = name
    if name in palettes:
        dict_name += " (stack palette)"
    palettes[dict_name] = StackPalette.load(name)
palettes = {k: palettes[k] for k in sorted(palettes)}
max_pal_size = max(len(pal) for pal in palettes.values())
carnival = StackPalette.load("carnival")
carnival_colorscale = Grad(carnival).to_plotly_colorscale()

fig = make_subplots(
    4, 2,
    specs=[[{"type": "scatter"}, {"type": "violin"}],
           [{"type": "scatter"}, {"type": "contour"}],
           [{"type": "histogram"}, {"type": "surface"}],
           [{"type": "choropleth", "colspan": 2}, None]]
)

# We will refer to this list to determine how to restyle the trace when changing palettes
restyle = []
for i in range(max_pal_size):
    y = [i]
    for j in range(49):
        y.append(y[-1] + 2 * (np.random.random() - 0.5))
    fig.add_trace(go.Scatter(
        x=np.arange(50),
        y=y,
        line_width=1,
        marker_size=10,
        line_color=carnival[i % len(carnival)],
        marker_color=carnival[i % len(carnival)],
        showlegend=False,
        visible=i < len(carnival)
    ), row=1, col=1)
    restyle.append("dynamic_categ")

    fig.add_trace(go.Violin(
        x=np.random.standard_normal(20) - i / 2,
        y0=max_pal_size - i,
        orientation='h',
        side="positive",
        width=3,
        line_color=carnival[(max_pal_size - i) % len(carnival)],
        showlegend=False,
        visible=i < len(carnival),
        hoverinfo="skip"
    ), row=1, col=2)
    restyle.append("dynamic_categ")

locations = []
pop = []
lifeexp = []
gdppercapita = []
with open(Path(__file__).parent / "gapminder.csv") as file:
    for row in csv.DictReader(file.read().split("\n")[:-1]):
        locations.append(row["iso"])
        pop.append(int(row["pop"]))
        lifeexp.append(float(row["lifeExp"]))
        gdppercapita.append(float(row["gdpPercap"]))
fig.add_trace(go.Scatter(
    x=gdppercapita,
    y=lifeexp,
    mode="markers",
    marker=go.scatter.Marker(
        size=pop,
        sizemode="area",
        sizeref=max(pop) / 3600,
        color=np.random.choice(np.array(carnival), size=len(pop))
    ),
    showlegend=False
), row=2, col=1)
restyle.append("static_categ")
fig.update_xaxes(type="log", row=2, col=1)

fig.add_trace(go.Contour(
    z=[[None, None, None, 12, 13, 14, 15, 16],
       [None, 1, None, 11, None, None, None, 17],
       [None, 2, 6, 7, None, None, None, 18],
       [None, 3, None, 8, None, None, None, 19],
       [5, 4, 10, 9, None, None, None, 20],
       [None, None, None, 27, None, None, None, 21],
       [None, None, None, 26, 25, 24, 23, 22]],
    connectgaps=True,
    colorscale=carnival_colorscale,
    showscale=False,
), row=2, col=2)
restyle.append("colorscale")

fig.add_trace(go.Histogram2d(
    x=np.concatenate([np.random.beta(4, 2, 100000), np.random.beta(2, 4, 100000)]),
    y=np.concatenate([np.random.beta(4, 2, 100000), np.random.beta(2, 4, 100000)]),
    nbinsx=50,
    nbinsy=50,
    colorscale=carnival_colorscale,
    showscale=False
), row=3, col=1)
restyle.append("colorscale")
fig.update_xaxes(constrain="domain", row=3, col=1)
fig.update_yaxes(scaleanchor="x5", scaleratio=1, row=3, col=1)


with open(Path(__file__).parent / "surface_data.csv") as file:
    surface_data = list(csv.reader(file.read().split("\n")[:-1]))
surface_data = np.array(surface_data, dtype=float)
fig.add_trace(go.Surface(
    z=surface_data,
    contours=dict(z=dict(show=True, usecolormap=True)),
    colorscale=carnival_colorscale,
    showscale=False
), row=3, col=2)
fig.update_layout(scene=dict(
    xaxis=dict(showticklabels=False, title=""),
    yaxis=dict(showticklabels=False, title=""),
    zaxis=dict(showticklabels=False, title="")
))
restyle.append("colorscale")

fig.add_trace(go.Choropleth(
    locations=locations,
    z=gdppercapita,
    colorscale=carnival_colorscale,
    colorbar=go.choropleth.ColorBar(x=1.1, y=1, yanchor="top", tickvals=[], len=0.4)
), row=4, col=1)
restyle.append("colorscale")

# This is just a dummy trace to display the discrete palette as a colorbar
fig.add_trace(go.Scatter(
    x=[None],
    y=[None],
    mode='markers',
    marker_colorscale=Grad(carnival, discrete=True).to_plotly_colorscale(),
    marker_showscale=True,
    marker_colorbar=go.scatter.marker.ColorBar(y=1, yanchor="top", tickvals=[], len=0.4),
    showlegend=False
))
restyle.append("discrete_colorscale")

fig.update_xaxes(showticklabels=False)
fig.update_yaxes(showticklabels=False)
fig.update_layout(mapbox_zoom=8,
                  width=700,
                  height=1300)

buttons = []
for pal_name, pal in palettes.items():
    restyle_dict = {
        "marker.color": [None] * len(fig.data),
        "marker.colorscale": [None] * len(fig.data),
        "marker.colorbar": [None] * len(fig.data),
        "line.color": [None] * len(fig.data),
        "colorscale": [None] * len(fig.data),
        "visible": [True] * len(fig.data)
    }
    for i, trace in enumerate(fig.data):
        if restyle[i] == "dynamic_categ":
            color_index = int(i / 2)
            if color_index >= len(pal):
                restyle_dict["visible"][i] = False
            else:
                restyle_dict["marker.color"][i] = pal[color_index]
                restyle_dict["line.color"][i] = pal[color_index]
        elif restyle[i] == "static_categ":
            restyle_dict["marker.color"][i] = np.random.choice(np.array(pal), size=len(pop))
        elif restyle[i] == "colorscale":
            restyle_dict["colorscale"][i] = Grad(pal).to_plotly_colorscale()
        elif restyle[i] == "discrete_colorscale":
            restyle_dict["marker.colorscale"][i] = Grad(pal, discrete=True).to_plotly_colorscale()
            restyle_dict["marker.colorbar"][i] = go.scatter.marker.ColorBar(y=1, yanchor="top", tickvals=[], len=0.4)
    buttons.append(go.layout.updatemenu.Button(
        label=pal_name,
        method="restyle",
        args=[restyle_dict]
    ))

fig.update_layout(
    updatemenus=[
        go.layout.Updatemenu(
            buttons=buttons,
            type='dropdown',
            direction='down',
            x=0.4, y=1.1,
            showactive=True,
            active=list(palettes).index("carnival")
        ),
        go.layout.Updatemenu(
            buttons=[
                go.layout.updatemenu.Button(
                    # We have to index with the key because of some lazy access magic plotly implements
                    args=[{"template": plotly_templates[template]}],
                    label=template,
                    method="relayout"
                ) for template in sorted(plotly_templates)
            ],
            type='dropdown',
            direction='down',
            x=0.8, y=1.1,
            showactive=True,
            active=sorted(plotly_templates).index("plotly")
        ),
    ],
    annotations=[
        go.layout.Annotation(
            text="Palette:",
            font=dict(size=12),
            x=0,
            y=1.1,
            xref="paper",
            yref="paper",
            showarrow=False
        ),
        go.layout.Annotation(
            text="Template:",
            font=dict(size=12),
            x=0.55,
            y=1.1,
            xanchor="right",
            xref="paper",
            yref="paper",
            showarrow=False
        ),
    ]
)
fig.show("browser")



# -*- coding: utf-8 -*-
import logging

import numpy
from numpy.linalg import pinv

from . import color_constants

logger = logging.getLogger(__name__)


# noinspection PyPep8Naming
def _get_adaptation_matrix(wp_src, wp_dst, observer, adaptation):
    """
    Calculate the correct transformation matrix based on origin and target
    illuminants. The observer angle must be the same between illuminants.

    See colormath.color_constants.ADAPTATION_MATRICES for a list of possible
    adaptations.

    Detailed conversion documentation is available at:
    http://brucelindbloom.com/Eqn_ChromAdapt.html
    """
    # Get the appropriate transformation matrix, [MsubA].
    m_sharp = color_constants.ADAPTATION_MATRICES[adaptation]

    # In case the white-points are still input as strings
    # Get white-points for illuminant
    if isinstance(wp_src, str):
        orig_illum = wp_src.lower()
        wp_src = color_constants.ILLUMINANTS[observer][orig_illum]
    elif hasattr(wp_src, "__iter__"):
        wp_src = wp_src

    if isinstance(wp_dst, str):
        targ_illum = wp_dst.lower()
        wp_dst = color_constants.ILLUMINANTS[observer][targ_illum]
    elif hasattr(wp_dst, "__iter__"):
        wp_dst = wp_dst

    # Sharpened cone responses ~ rho gamma beta ~ sharpened r g b
    rgb_src = numpy.dot(m_sharp, wp_src)
    rgb_dst = numpy.dot(m_sharp, wp_dst)

    # Ratio of whitepoint sharpened responses
    m_rat = numpy.diag(rgb_dst / rgb_src)

    # Final transformation matrix
    m_xfm = numpy.dot(numpy.dot(pinv(m_sharp), m_rat), m_sharp)

    return m_xfm


# noinspection PyPep8Naming
def apply_chromatic_adaptation(
    val_x, val_y, val_z, orig_illum, targ_illum, observer="2", adaptation="bradford"
):
    """
    Applies a chromatic adaptation matrix to convert XYZ values between
    illuminants. It is important to recognize that color transformation results
    in color errors, determined by how far the original illuminant is from the
    target illuminant. For example, D65 to A could result in very high maximum
    deviance.

    An informative article with estimate average Delta E values for each
    illuminant conversion may be found at:

    http://brucelindbloom.com/ChromAdaptEval.html
    """
    # It's silly to have to do this, but some people may want to call this
    # function directly, so we'll protect them from messing up upper/lower case.
    adaptation = adaptation.lower()

    # Get white-points for illuminant
    if isinstance(orig_illum, str):
        orig_illum = orig_illum.lower()
        wp_src = color_constants.ILLUMINANTS[observer][orig_illum]
    elif hasattr(orig_illum, "__iter__"):
        wp_src = orig_illum

    if isinstance(targ_illum, str):
        targ_illum = targ_illum.lower()
        wp_dst = color_constants.ILLUMINANTS[observer][targ_illum]
    elif hasattr(targ_illum, "__iter__"):
        wp_dst = targ_illum

    logger.debug("  \\* Applying adaptation matrix: %s", adaptation)
    # Retrieve the appropriate transformation matrix from the constants.
    transform_matrix = _get_adaptation_matrix(wp_src, wp_dst, observer, adaptation)

    # Stuff the XYZ values into a NumPy matrix for conversion.
    XYZ_matrix = numpy.array((val_x, val_y, val_z))
    # Perform the adaptation via matrix multiplication.
    result_matrix = numpy.dot(transform_matrix, XYZ_matrix)

    # Return individual X, Y, and Z coordinates.
    return result_matrix[0], result_matrix[1], result_matrix[2]


# noinspection PyPep8Naming
def apply_chromatic_adaptation_on_color(color, targ_illum, adaptation="bradford"):
    """
    Convenience function to apply an adaptation directly to a Color object.
    """
    xyz_x = color.xyz_x
    xyz_y = color.xyz_y
    xyz_z = color.xyz_z
    orig_illum = color.illuminant
    targ_illum = targ_illum.lower()
    observer = color.observer
    adaptation = adaptation.lower()

    # Return individual X, Y, and Z coordinates.
    color.xyz_x, color.xyz_y, color.xyz_z = apply_chromatic_adaptation(
        xyz_x,
        xyz_y,
        xyz_z,
        orig_illum,
        targ_illum,
        observer=observer,
        adaptation=adaptation,
    )
    color.set_illuminant(targ_illum)

    return color



# -*- coding: utf-8 -*-
"""
This module contains classes to represent various color spaces.
"""

import logging
import math

import numpy

from colorir.colormath import color_constants
from colorir.colormath import density
from colorir.colormath.chromatic_adaptation import apply_chromatic_adaptation_on_color
from colorir.colormath.color_exceptions import InvalidObserverError, InvalidIlluminantError

logger = logging.getLogger(__name__)


class ColorBase(object):
    """
    A base class holding some common methods and values.
    """

    # Attribute names containing color data on the sub-class. For example,
    # sRGBColor would be ['rgb_r', 'rgb_g', 'rgb_b']
    VALUES = []
    # If this object as converted such that its values passed through an
    # RGB colorspace, this is set to the class for said RGB color space.
    # Allows reversing conversions automatically and accurately.
    _through_rgb_type = None

    def get_value_tuple(self):
        """
        Returns a tuple of the color's values (in order). For example,
        an LabColor object will return (lab_l, lab_a, lab_b), where each
        member of the tuple is the float value for said variable.
        """
        retval = tuple()
        for val in self.VALUES:
            retval += (getattr(self, val),)
        return retval

    def __str__(self):
        """
        String representation of the color.
        """
        retval = self.__class__.__name__ + " ("
        for val in self.VALUES:
            value = getattr(self, val, None)
            if value is not None:
                retval += "%s:%.4f " % (val, getattr(self, val))
        if hasattr(self, "observer"):
            retval += "observer:" + self.observer
        if hasattr(self, "illuminant"):
            retval += " illuminant:" + self.illuminant
        return retval.strip() + ")"

    def __repr__(self):
        """
        Evaluable string representation of the object.
        """
        retval = self.__class__.__name__ + "("
        attributes = [(attr, getattr(self, attr)) for attr in self.VALUES]
        values = [x + "=" + repr(y) for x, y in attributes]
        retval += ", ".join(values)
        if hasattr(self, "observer"):
            retval += ", observer='" + self.observer + "'"
        if hasattr(self, "illuminant"):
            retval += ", illuminant='" + self.illuminant + "'"
        return retval + ")"


class IlluminantMixin(object):
    """
    Color spaces that have a notion of an illuminant should inherit this.
    """

    # noinspection PyAttributeOutsideInit
    def set_observer(self, observer):
        """
        Validates and sets the color's observer angle.

        .. note:: This only changes the observer angle value. It does no conversion
            of the color's coordinates.

        :param str observer: One of '2' or '10'.
        """
        observer = str(observer)
        if observer not in color_constants.OBSERVERS:
            raise InvalidObserverError(self)
        self.observer = observer

    # noinspection PyAttributeOutsideInit
    def set_illuminant(self, illuminant):
        """
        Validates and sets the color's illuminant.

        .. note:: This only changes the illuminant. It does no conversion
            of the color's coordinates. For this, you'll want to refer to
            :py:meth:`XYZColor.apply_adaptation \
<colormath.color_objects.XYZColor.apply_adaptation>`.

        .. tip:: Call this after setting your observer.

        :param str illuminant: One of the various illuminants.
        """
        illuminant = illuminant.lower()
        if illuminant not in color_constants.ILLUMINANTS[self.observer]:
            raise InvalidIlluminantError(illuminant)
        self.illuminant = illuminant

    def get_illuminant_xyz(self, observer=None, illuminant=None):
        """
        :param str observer: Get the XYZ values for another observer angle. Must
            be either '2' or '10'.
        :param str illuminant: Get the XYZ values for another illuminant.
        :returns: the color's illuminant's XYZ values.
        """
        try:
            if observer is None:
                observer = self.observer

            illums_observer = color_constants.ILLUMINANTS[observer]
        except KeyError:
            raise InvalidObserverError(self)

        try:
            if illuminant is None:
                illuminant = self.illuminant

            illum_xyz = illums_observer[illuminant]
        except (KeyError, AttributeError):
            raise InvalidIlluminantError(illuminant)

        return {"X": illum_xyz[0], "Y": illum_xyz[1], "Z": illum_xyz[2]}


class SpectralColor(IlluminantMixin, ColorBase):
    """
    A SpectralColor represents a spectral power distribution, as read by
    a spectrophotometer. Our current implementation has wavelength intervals
    of 10nm, starting at 340nm and ending at 830nm.

    Spectral colors are the lowest level, most "raw" measurement of color.
    You may convert spectral colors to any other color space, but you can't
    convert any other color space back to spectral.

    See `Spectral power distribution \
<http://en.wikipedia.org/wiki/Spectral_power_distribution>`_
    on Wikipedia for some higher level details on how these work.
    """

    VALUES = [
        "spec_340nm",
        "spec_350nm",
        "spec_360nm",
        "spec_370nm",
        "spec_380nm",
        "spec_390nm",
        "spec_400nm",
        "spec_410nm",
        "spec_420nm",
        "spec_430nm",
        "spec_440nm",
        "spec_450nm",
        "spec_460nm",
        "spec_470nm",
        "spec_480nm",
        "spec_490nm",
        "spec_500nm",
        "spec_510nm",
        "spec_520nm",
        "spec_530nm",
        "spec_540nm",
        "spec_550nm",
        "spec_560nm",
        "spec_570nm",
        "spec_580nm",
        "spec_590nm",
        "spec_600nm",
        "spec_610nm",
        "spec_620nm",
        "spec_630nm",
        "spec_640nm",
        "spec_650nm",
        "spec_660nm",
        "spec_670nm",
        "spec_680nm",
        "spec_690nm",
        "spec_700nm",
        "spec_710nm",
        "spec_720nm",
        "spec_730nm",
        "spec_740nm",
        "spec_750nm",
        "spec_760nm",
        "spec_770nm",
        "spec_780nm",
        "spec_790nm",
        "spec_800nm",
        "spec_810nm",
        "spec_820nm",
        "spec_830nm",
    ]

    def __init__(
        self,
        spec_340nm=0.0,
        spec_350nm=0.0,
        spec_360nm=0.0,
        spec_370nm=0.0,
        spec_380nm=0.0,
        spec_390nm=0.0,
        spec_400nm=0.0,
        spec_410nm=0.0,
        spec_420nm=0.0,
        spec_430nm=0.0,
        spec_440nm=0.0,
        spec_450nm=0.0,
        spec_460nm=0.0,
        spec_470nm=0.0,
        spec_480nm=0.0,
        spec_490nm=0.0,
        spec_500nm=0.0,
        spec_510nm=0.0,
        spec_520nm=0.0,
        spec_530nm=0.0,
        spec_540nm=0.0,
        spec_550nm=0.0,
        spec_560nm=0.0,
        spec_570nm=0.0,
        spec_580nm=0.0,
        spec_590nm=0.0,
        spec_600nm=0.0,
        spec_610nm=0.0,
        spec_620nm=0.0,
        spec_630nm=0.0,
        spec_640nm=0.0,
        spec_650nm=0.0,
        spec_660nm=0.0,
        spec_670nm=0.0,
        spec_680nm=0.0,
        spec_690nm=0.0,
        spec_700nm=0.0,
        spec_710nm=0.0,
        spec_720nm=0.0,
        spec_730nm=0.0,
        spec_740nm=0.0,
        spec_750nm=0.0,
        spec_760nm=0.0,
        spec_770nm=0.0,
        spec_780nm=0.0,
        spec_790nm=0.0,
        spec_800nm=0.0,
        spec_810nm=0.0,
        spec_820nm=0.0,
        spec_830nm=0.0,
        observer="2",
        illuminant="d50",
    ):
        """
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """

        super(SpectralColor, self).__init__()
        # Spectral fields
        self.spec_340nm = float(spec_340nm)
        self.spec_350nm = float(spec_350nm)
        self.spec_360nm = float(spec_360nm)
        self.spec_370nm = float(spec_370nm)
        # begin Blue wavelengths
        self.spec_380nm = float(spec_380nm)
        self.spec_390nm = float(spec_390nm)
        self.spec_400nm = float(spec_400nm)
        self.spec_410nm = float(spec_410nm)
        self.spec_420nm = float(spec_420nm)
        self.spec_430nm = float(spec_430nm)
        self.spec_440nm = float(spec_440nm)
        self.spec_450nm = float(spec_450nm)
        self.spec_460nm = float(spec_460nm)
        self.spec_470nm = float(spec_470nm)
        self.spec_480nm = float(spec_480nm)
        self.spec_490nm = float(spec_490nm)
        # end Blue wavelengths
        # start Green wavelengths
        self.spec_500nm = float(spec_500nm)
        self.spec_510nm = float(spec_510nm)
        self.spec_520nm = float(spec_520nm)
        self.spec_530nm = float(spec_530nm)
        self.spec_540nm = float(spec_540nm)
        self.spec_550nm = float(spec_550nm)
        self.spec_560nm = float(spec_560nm)
        self.spec_570nm = float(spec_570nm)
        self.spec_580nm = float(spec_580nm)
        self.spec_590nm = float(spec_590nm)
        self.spec_600nm = float(spec_600nm)
        self.spec_610nm = float(spec_610nm)
        # end Green wavelengths
        # start Red wavelengths
        self.spec_620nm = float(spec_620nm)
        self.spec_630nm = float(spec_630nm)
        self.spec_640nm = float(spec_640nm)
        self.spec_650nm = float(spec_650nm)
        self.spec_660nm = float(spec_660nm)
        self.spec_670nm = float(spec_670nm)
        self.spec_680nm = float(spec_680nm)
        self.spec_690nm = float(spec_690nm)
        self.spec_700nm = float(spec_700nm)
        self.spec_710nm = float(spec_710nm)
        self.spec_720nm = float(spec_720nm)
        # end Red wavelengths
        self.spec_730nm = float(spec_730nm)
        self.spec_740nm = float(spec_740nm)
        self.spec_750nm = float(spec_750nm)
        self.spec_760nm = float(spec_760nm)
        self.spec_770nm = float(spec_770nm)
        self.spec_780nm = float(spec_780nm)
        self.spec_790nm = float(spec_790nm)
        self.spec_800nm = float(spec_800nm)
        self.spec_810nm = float(spec_810nm)
        self.spec_820nm = float(spec_820nm)
        self.spec_830nm = float(spec_830nm)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)

    def get_numpy_array(self):
        """
        Dump this color into NumPy array.
        """
        # This holds the obect's spectral data, and will be passed to
        # numpy.array() to create a numpy array (matrix) for the matrix math
        # that will be done during the conversion to XYZ.
        values = []

        # Use the required value list to build this dynamically. Default to
        # 0.0, since that ultimately won't affect the outcome due to the math
        # involved.
        for val in self.VALUES:
            values.append(getattr(self, val, 0.0))

        # Create and the actual numpy array/matrix from the spectral list.
        color_array = numpy.array([values])
        return color_array

    def calc_density(self, density_standard=None):
        """
        Calculates the density of the SpectralColor. By default, Status T
        density is used, and the correct density distribution (Red, Green,
        or Blue) is chosen by comparing the Red, Green, and Blue components of
        the spectral sample (the values being red in via "filters").
        """
        if density_standard is not None:
            return density.ansi_density(self, density_standard)
        else:
            return density.auto_density(self)


class LabColor(IlluminantMixin, ColorBase):
    """
    Represents a CIE Lab color. For more information on CIE Lab,
    see `Lab color space <http://en.wikipedia.org/wiki/Lab_color_space>`_ on
    Wikipedia.
    """

    VALUES = ["lab_l", "lab_a", "lab_b"]

    def __init__(self, lab_l, lab_a, lab_b, observer="2", illuminant="d50"):
        """
        :param float lab_l: L coordinate.
        :param float lab_a: a coordinate.
        :param float lab_b: b coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(LabColor, self).__init__()
        #: L coordinate
        self.lab_l = float(lab_l)
        #: a coordinate
        self.lab_a = float(lab_a)
        #: b coordinate
        self.lab_b = float(lab_b)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)


class LCHabColor(IlluminantMixin, ColorBase):
    """
    Represents an CIE LCH color that was converted to LCH by passing through
    CIE Lab.  This differs from :py:class:`LCHuvColor`, which was converted to
    LCH through CIE Luv.

    See `Introduction to Colour Spaces \
<http://www.colourphil.co.uk/lab_lch_colour_space.shtml>`_ by Phil Cruse for an
    illustration of how CIE LCH differs from CIE Lab.
    """

    VALUES = ["lch_l", "lch_c", "lch_h"]

    def __init__(self, lch_l, lch_c, lch_h, observer="2", illuminant="d50"):
        """
        :param float lch_l: L coordinate.
        :param float lch_c: C coordinate.
        :param float lch_h: H coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(LCHabColor, self).__init__()
        #: L coordinate
        self.lch_l = float(lch_l)
        #: C coordinate
        self.lch_c = float(lch_c)
        #: H coordinate
        self.lch_h = float(lch_h)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)


class LCHuvColor(IlluminantMixin, ColorBase):
    """
    Represents an CIE LCH color that was converted to LCH by passing through
    CIE Luv.  This differs from :py:class:`LCHabColor`, which was converted to
    LCH through CIE Lab.

    See `Introduction to Colour Spaces \
<http://www.colourphil.co.uk/lab_lch_colour_space.shtml>`_ by Phil Cruse for an
    illustration of how CIE LCH differs from CIE Lab.
    """

    VALUES = ["lch_l", "lch_c", "lch_h"]

    def __init__(self, lch_l, lch_c, lch_h, observer="2", illuminant="d50"):
        """
        :param float lch_l: L coordinate.
        :param float lch_c: C coordinate.
        :param float lch_h: H coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(LCHuvColor, self).__init__()
        #: L coordinate
        self.lch_l = float(lch_l)
        #: C coordinate
        self.lch_c = float(lch_c)
        #: H coordinate
        self.lch_h = float(lch_h)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)


class LuvColor(IlluminantMixin, ColorBase):
    """
    Represents an Luv color.
    """

    VALUES = ["luv_l", "luv_u", "luv_v"]

    def __init__(self, luv_l, luv_u, luv_v, observer="2", illuminant="d50"):
        """
        :param float luv_l: L coordinate.
        :param float luv_u: u coordinate.
        :param float luv_v: v coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(LuvColor, self).__init__()
        #: L coordinate
        self.luv_l = float(luv_l)
        #: u coordinate
        self.luv_u = float(luv_u)
        #: v coordinate
        self.luv_v = float(luv_v)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)


class XYZColor(IlluminantMixin, ColorBase):
    """
    Represents an XYZ color.
    """

    VALUES = ["xyz_x", "xyz_y", "xyz_z"]

    def __init__(self, xyz_x, xyz_y, xyz_z, observer="2", illuminant="d50"):
        """
        :param float xyz_x: X coordinate.
        :param float xyz_y: Y coordinate.
        :param float xyz_z: Z coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(XYZColor, self).__init__()
        #: X coordinate
        self.xyz_x = float(xyz_x)
        #: Y coordinate
        self.xyz_y = float(xyz_y)
        #: Z coordinate
        self.xyz_z = float(xyz_z)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)

    def apply_adaptation(self, target_illuminant, adaptation="bradford"):
        """
        This applies an adaptation matrix to change the XYZ color's illuminant.
        You'll most likely only need this during RGB conversions.
        """
        logger.debug("  \\- Original illuminant: %s", self.illuminant)
        logger.debug("  \\- Target illuminant: %s", target_illuminant)

        # If the XYZ values were taken with a different reference white than the
        # native reference white of the target RGB space, a transformation matrix
        # must be applied.
        if self.illuminant != target_illuminant:
            logger.debug(
                "  \\* Applying transformation from %s to %s ",
                self.illuminant,
                target_illuminant,
            )
            # Sets the adjusted XYZ values, and the new illuminant.
            apply_chromatic_adaptation_on_color(
                color=self, targ_illum=target_illuminant, adaptation=adaptation
            )


# noinspection PyPep8Naming
class xyYColor(IlluminantMixin, ColorBase):
    """
    Represents an xyY color.
    """

    VALUES = ["xyy_x", "xyy_y", "xyy_Y"]

    def __init__(self, xyy_x, xyy_y, xyy_Y, observer="2", illuminant="d50"):
        """
        :param float xyy_x: x coordinate.
        :param float xyy_y: y coordinate.
        :param float xyy_Y: Y coordinate.
        :keyword str observer: Observer angle. Either ``'2'`` or ``'10'`` degrees.
        :keyword str illuminant: See :doc:`illuminants` for valid values.
        """
        super(xyYColor, self).__init__()
        #: x coordinate
        self.xyy_x = float(xyy_x)
        #: y coordinate
        self.xyy_y = float(xyy_y)
        #: Y coordinate
        self.xyy_Y = float(xyy_Y)

        #: The color's observer angle. Set with :py:meth:`set_observer`.
        self.observer = None
        #: The color's illuminant. Set with :py:meth:`set_illuminant`.
        self.illuminant = None

        self.set_observer(observer)
        self.set_illuminant(illuminant)


class BaseRGBColor(ColorBase):
    """
    Base class for all RGB color spaces.

    .. warning:: Do not use this class directly!
    """

    VALUES = ["rgb_r", "rgb_g", "rgb_b"]

    def __init__(self, rgb_r, rgb_g, rgb_b, is_upscaled=False):
        """
        :param float rgb_r: R coordinate. 0.0-1.0, or 0-255 if is_upscaled=True.
        :param float rgb_g: G coordinate. 0.0-1.0, or 0-255 if is_upscaled=True.
        :param float rgb_b: B coordinate. 0.0-1.0, or 0-255 if is_upscaled=True.
        :keyword bool is_upscaled: If False, RGB coordinate values are
            between 0.0 and 1.0. If True, RGB values are between 0 and 255.
        """
        super(BaseRGBColor, self).__init__()
        if is_upscaled:
            self.rgb_r = rgb_r / 255.0
            self.rgb_g = rgb_g / 255.0
            self.rgb_b = rgb_b / 255.0
        else:
            self.rgb_r = float(rgb_r)
            self.rgb_g = float(rgb_g)
            self.rgb_b = float(rgb_b)
        self.is_upscaled = is_upscaled

    def _clamp_rgb_coordinate(self, coord):
        """
        Clamps an RGB coordinate, taking into account whether or not the
        color is upscaled or not.

        :param float coord: The coordinate value.
        :rtype: float
        :returns: The clamped value.
        """
        if not self.is_upscaled:
            return min(max(coord, 0.0), 1.0)
        else:
            return min(max(coord, 0.0), 255.0)

    @property
    def clamped_rgb_r(self):
        """
        The clamped (0.0-1.0) R value.
        """
        return self._clamp_rgb_coordinate(self.rgb_r)

    @property
    def clamped_rgb_g(self):
        """
        The clamped (0.0-1.0) G value.
        """
        return self._clamp_rgb_coordinate(self.rgb_g)

    @property
    def clamped_rgb_b(self):
        """
        The clamped (0.0-1.0) B value.
        """
        return self._clamp_rgb_coordinate(self.rgb_b)

    def get_upscaled_value_tuple(self):
        """
        Scales an RGB color object from decimal 0.0-1.0 to int 0-255.
        """
        # Scale up to 0-255 values.
        rgb_r = int(math.floor(0.5 + self.rgb_r * 255))
        rgb_g = int(math.floor(0.5 + self.rgb_g * 255))
        rgb_b = int(math.floor(0.5 + self.rgb_b * 255))

        return rgb_r, rgb_g, rgb_b

    def get_rgb_hex(self):
        """
        Converts the RGB value to a hex value in the form of: #RRGGBB

        :rtype: str
        """
        rgb_r, rgb_g, rgb_b = self.get_upscaled_value_tuple()
        return "#%02x%02x%02x" % (rgb_r, rgb_g, rgb_b)

    @classmethod
    def new_from_rgb_hex(cls, hex_str):
        """
        Converts an RGB hex string like #RRGGBB and assigns the values to
        this sRGBColor object.

        :rtype: sRGBColor
        """
        colorstring = hex_str.strip()
        if colorstring[0] == "#":
            colorstring = colorstring[1:]
        if len(colorstring) != 6:
            raise ValueError("input #%s is not in #RRGGBB format" % colorstring)
        r, g, b = colorstring[:2], colorstring[2:4], colorstring[4:]
        r, g, b = [int(n, 16) / 255.0 for n in (r, g, b)]
        return cls(r, g, b)


# noinspection PyPep8Naming
class sRGBColor(BaseRGBColor):
    """
    Represents an sRGB color.

    .. note:: If you pass in upscaled values, we automatically scale them
        down to 0.0-1.0. If you need the old upscaled values, you can
        retrieve them with :py:meth:`get_upscaled_value_tuple`.

    :ivar float rgb_r: R coordinate
    :ivar float rgb_g: G coordinate
    :ivar float rgb_b: B coordinate
    :ivar bool is_upscaled: If True, RGB values are between 0-255. If False,
        0.0-1.0.
    """

    #: RGB space's gamma constant.
    rgb_gamma = 2.2
    #: The RGB space's native illuminant. Important when converting to XYZ.
    native_illuminant = "d65"
    conversion_matrices = {
        "xyz_to_rgb": numpy.array(
            (
                (3.24071, -1.53726, -0.498571),
                (-0.969258, 1.87599, 0.0415557),
                (0.0556352, -0.203996, 1.05707),
            )
        ),
        "rgb_to_xyz": numpy.array(
            (
                (0.412424, 0.357579, 0.180464),
                (0.212656, 0.715158, 0.0721856),
                (0.0193324, 0.119193, 0.950444),
            )
        ),
    }


class BT2020Color(BaseRGBColor):
    """
    Represents a ITU-R BT.2020 color.

    .. note:: If you pass in upscaled values, we automatically scale them
        down to 0.0-1.0. If you need the old upscaled values, you can
        retrieve them with :py:meth:`get_upscaled_value_tuple`.

    :ivar float rgb_r: R coordinate
    :ivar float rgb_g: G coordinate
    :ivar float rgb_b: B coordinate
    :ivar bool is_upscaled: If True, RGB values are between 0-255. If False,
        0.0-1.0.
    """

    #: RGB space's gamma constant.
    rgb_gamma = 2.4
    #: The RGB space's native illuminant. Important when converting to XYZ.
    native_illuminant = "d65"
    conversion_matrices = {
        "xyz_to_rgb": numpy.array(
            (
                (1.716651187971269, -0.355670783776393, -0.253366281373660),
                (-0.666684351832489, 1.616481236634939, 0.015768545813911),
                (0.017639857445311, -0.042770613257809, 0.942103121235474),
            )
        ),
        "rgb_to_xyz": numpy.array(
            (
                (0.636958048301291, 0.144616903586208, 0.168880975164172),
                (0.262700212011267, 0.677998071518871, 0.059301716469862),
                (0.000000000000000, 0.028072693049087, 1.060985057710791),
            )
        ),
    }


class AdobeRGBColor(BaseRGBColor):
    """
    Represents an Adobe RGB color.

    .. note:: If you pass in upscaled values, we automatically scale them
        down to 0.0-1.0. If you need the old upscaled values, you can
        retrieve them with :py:meth:`get_upscaled_value_tuple`.

    :ivar float rgb_r: R coordinate
    :ivar float rgb_g: G coordinate
    :ivar float rgb_b: B coordinate
    :ivar bool is_upscaled: If True, RGB values are between 0-255. If False,
        0.0-1.0.
    """

    #: RGB space's gamma constant.
    rgb_gamma = 2.2
    #: The RGB space's native illuminant. Important when converting to XYZ.
    native_illuminant = "d65"
    conversion_matrices = {
        "xyz_to_rgb": numpy.array(
            (
                (2.04148, -0.564977, -0.344713),
                (-0.969258, 1.87599, 0.0415557),
                (0.0134455, -0.118373, 1.01527),
            )
        ),
        "rgb_to_xyz": numpy.array(
            (
                (0.576700, 0.185556, 0.188212),
                (0.297361, 0.627355, 0.0752847),
                (0.0270328, 0.0706879, 0.991248),
            )
        ),
    }


class AppleRGBColor(BaseRGBColor):
    """
    Represents an AppleRGB color.

    .. note:: If you pass in upscaled values, we automatically scale them
        down to 0.0-1.0. If you need the old upscaled values, you can
        retrieve them with :py:meth:`get_upscaled_value_tuple`.

    :ivar float rgb_r: R coordinate
    :ivar float rgb_g: G coordinate
    :ivar float rgb_b: B coordinate
    :ivar bool is_upscaled: If True, RGB values are between 0-255. If False,
        0.0-1.0.
    """

    #: RGB space's gamma constant.
    rgb_gamma = 1.8
    #: The RGB space's native illuminant. Important when converting to XYZ.
    native_illuminant = "d65"
    conversion_matrices = {
        "xyz_to_rgb": numpy.array(
            (
                (2.9515373, -1.2894116, -0.4738445),
                (-1.0851093, 1.9908566, 0.0372026),
                (0.0854934, -0.2694964, 1.0912975),
            )
        ),
        "rgb_to_xyz": numpy.array(
            (
                (0.4497288, 0.3162486, 0.1844926),
                (0.2446525, 0.6720283, 0.0833192),
                (0.0251848, 0.1411824, 0.9224628),
            )
        ),
    }


class HSLColor(ColorBase):
    """
    Represents an HSL color.
    """

    VALUES = ["hsl_h", "hsl_s", "hsl_l"]

    def __init__(self, hsl_h, hsl_s, hsl_l):
        """
        :param float hsl_h: H coordinate.
        :param float hsl_s: S coordinate.
        :param float hsl_l: L coordinate.
        """
        super(HSLColor, self).__init__()
        #: H coordinate
        self.hsl_h = float(hsl_h)
        #: S coordinate
        self.hsl_s = float(hsl_s)
        #: L coordinate
        self.hsl_l = float(hsl_l)


class HSVColor(ColorBase):
    """
    Represents an HSV color.
    """

    VALUES = ["hsv_h", "hsv_s", "hsv_v"]

    def __init__(self, hsv_h, hsv_s, hsv_v):
        """
        :param float hsv_h: H coordinate.
        :param float hsv_s: S coordinate.
        :param float hsv_v: V coordinate.
        """
        super(HSVColor, self).__init__()
        #: H coordinate
        self.hsv_h = float(hsv_h)
        #: S coordinate
        self.hsv_s = float(hsv_s)
        #: V coordinate
        self.hsv_v = float(hsv_v)


class CMYColor(ColorBase):
    """
    Represents a CMY color.
    """

    VALUES = ["cmy_c", "cmy_m", "cmy_y"]

    def __init__(self, cmy_c, cmy_m, cmy_y):
        """
        :param float cmy_c: C coordinate.
        :param float cmy_m: M coordinate.
        :param float cmy_y: Y coordinate.
        """
        super(CMYColor, self).__init__()
        #: C coordinate
        self.cmy_c = float(cmy_c)
        #: M coordinate
        self.cmy_m = float(cmy_m)
        #: Y coordinate
        self.cmy_y = float(cmy_y)


class CMYKColor(ColorBase):
    """
    Represents a CMYK color.
    """

    VALUES = ["cmyk_c", "cmyk_m", "cmyk_y", "cmyk_k"]

    def __init__(self, cmyk_c, cmyk_m, cmyk_y, cmyk_k):
        """
        :param float cmyk_c: C coordinate.
        :param float cmyk_m: M coordinate.
        :param float cmyk_y: Y coordinate.
        :param float cmyk_k: K coordinate.
        """
        super(CMYKColor, self).__init__()
        #: C coordinate
        self.cmyk_c = float(cmyk_c)
        #: M coordinate
        self.cmyk_m = float(cmyk_m)
        #: Y coordinate
        self.cmyk_y = float(cmyk_y)
        #: K coordinate
        self.cmyk_k = float(cmyk_k)


class IPTColor(ColorBase):
    """
    Represents an IPT color.

    Reference:
    Fairchild, M. D. (2013). Color appearance models, 3rd Ed. (pp. 271-272).
    John Wiley & Sons.
    """

    VALUES = ["ipt_i", "ipt_p", "ipt_t"]

    conversion_matrices = {
        "xyz_to_lms": numpy.array(
            (
                (0.4002, 0.7075, -0.0807),
                (-0.2280, 1.1500, 0.0612),
                (0.0000, 0.0000, 0.9184),
            )
        ),
        "lms_to_ipt": numpy.array(
            (
                (0.4000, 0.4000, 0.2000),
                (4.4550, -4.8510, 0.3960),
                (0.8056, 0.3572, -1.1628),
            )
        ),
    }

    def __init__(self, ipt_i, ipt_p, ipt_t):
        """
        :param ipt_i: I coordinate.
        :param ipt_p: P coordinate.
        :param ipt_t: T coordinate.
        """
        super(IPTColor, self).__init__()
        #: I coordinate
        self.ipt_i = ipt_i
        #: P coordinate
        self.ipt_p = ipt_p
        #: T coordinate
        self.ipt_t = ipt_t

    @property
    def hue_angle(self):
        return numpy.arctan2(self.ipt_t, self.ipt_p)
